/************************************************************************************************************************/
/*

Software for smart cameras from Vision Components
-------------------------------------------------

Program: Klaus Schneider, VC

VC 3D Laser Scanner Demo

The program runs as a standalone program on a VC62xx camera.
For 3D result display or parameter settings use the PC Host Software VC3D.exe.

Main functions are:

I32  ScannerInit       (LsParameter *LsPar, I32 Mode);
I32  ScannerDetection  (LsParameter *LsPar, image *ImgArea);
void ImgToWorldCoord   (LsParameter *LsPar, I32 Nr, float *restrict pImgX, float *restrict pImgZ, float *restrict pWorldX, float *restrict pWorldZ);

Call once at the beginning of the program the function ScannerInit(LsPar, 0) and at the end of your program the function ScannerInit(LsPar, 1)
in order to release memory.

The program is designed for the VCLIB 3.06 and higher.


Program Description:
--------------------



*/
/************************************************************************************************************************/

#include "LaserScanner_Header.h"

/************************************************************************************************************************/

#define MAIN_VERSION  49

#define LASER_POWER  100 // 30 or 100 (mW) laser power (you can use always 100 mW as laser is controlled by voltage)

// #define PRODUCT_VERSION

/***************************************************************************************************************************/

// TCP / IP
#define SERVER_PORT         1096    // scanner communication port
#define TCP_TIMEOUT          500    // ms

/************************************************************************************************************************/

#ifdef PRODUCT_VERSION
extern ProductParameter ProdPar[1];
#endif

/************************************************************************************************************************/

///////////////////////////////////////////////////////////////////////////////////////////////////////
// TCP / IP Functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

I32  TCP_RunCommand      (LsParameter *LsPar, U8 *Command, I32 length, U32 *Alias);
I32  TCP_SendCommand     (LsParameter *LsPar, U32 *Alias, U8 *Command, I32 maxlen, I32 BufferMode);

I32  TCP_ReadCommand     (U32 sock, U32 *Alias, char *Byte, I32 minlen, I32 maxlen, I32 *length);
I32  TCP_DataAvailable   (U32 sock, U32 *Alias, char *Byte, I32 length);
I32  TCP_Open            (I32 Port, U32 *NewSock);
I32  TCP_Connection      (U32 timeout, U32 *NewSock);
I32  TCP_Accept          (U32 Sock, U32 *FromSock);
I32  TCP_Send            (U32 Sock, char *SendData, I32 SendBytes);
I32  TCP_Receive         (U32 Sock, char *ReceiveData, I32 ReceiveBytes);
void TCP_Close           (U32 Sock);


///////////////////////////////////////////////////////////////////////////////////////////////////////
// Function declarations
///////////////////////////////////////////////////////////////////////////////////////////////////////

I32  IniProductParameter (LsParameter *LsPar, I32 Mode);
I32  SetProductParameter (LsParameter *LsPar);
I32  Product_Measure     (LsParameter *LsPar);

I32  InitSensorParameter (LsParameter *LsPar, I32 Mode);
I32  GetParameterRange   (LsParameter *LsPar, void *pPar, I32 *SetVal, I32 *MinVal, I32 *MaxVal);
I32  ReadIntParameter    (LsParameter *LsPar, I8 *pParFile, I32 *pParStruct, I32 Debug);
void ChangeParameter     (LsParameter *LsPar);
void ChangeValue_Sensor  (LsParameter *LsPar, I32 *ParValue, I32 Step, I8 *text);
void ChangeValue_Telnet  (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text);
int  ReadKeyboard_Telnet (int wait);
I32  ReadyForCalc        (LsParameter *LsPar);
I32  ScreenBuffer2       (LsParameter *LsPar, I32 Mode);
I32  AutoShutter         (LsParameter *LsPar);
I32  AutoTrigger         (LsParameter *LsPar, I32 AutoMode);
I32  SwitchLaser         (I32 LaserOn);
I32  SetImageFormat      (LsParameter *LsPar);
I32  StoreImage          (void);
I32  ReadImage           (void);

I32  set_image_format_ev76c560_subs (I32 subsampling_factor, I32 hpos, I32 vpos, I32 hsize, I32 vsize, I40 shutter, I32 mode, I32 * sen_hpos, I32 * sen_vpos, I32 * sen_hsize, I32 * sen_vsize, I40 * sen_shutter);

/************************************************************************************************************************/

// global values
U32 HostCounter = 0, CamCounter  = 0;
U32 LineCounter = 0;

/************************************************************************************************************************/

void main(void)
{
  LsParameter   LsPar[1];

  I32           response=-1, count, RunMode = 1, AutoTrigMode = 0, Error = 0;
  U32           sock=0, alias=0;
  I40           MsD, Ms0=-1, Ms1;


  /////////////////////////////////////////////////////////////////////////////////
  // set higher task priority
  /////////////////////////////////////////////////////////////////////////////////

  I32 oldprio, highprio = 8, id = task_get_id(); // valid priority range: 1=highest 18=lowest priority

  task_set_priority(id, highprio, &oldprio); // set task priority
  print("Set own task '%s' to prio %d! (oldprio=%d)\n", task_get_name_from_id(id), highprio, oldprio);
  

  /////////////////////////////////////////////////////////////////////////////////
  // init all VC library licences
  /////////////////////////////////////////////////////////////////////////////////

  /*******************************************************************************/
  /*                 PLEASE INITIALIZE YOUR LICENCE CODES                        */
  /*                    to activate library functions                            */
  /*                                                                             */
  /* VCLIB :         init_licence("T1122334455") or init_licence("LT1122334455") */
  /* Extension Lib : init_licence("E1122334455") or init_licence("LE1122334455") */
  /* Color Lib :     init_licence("C1122334455") or init_licence("LC1122334455") */
  /* etc.                                                                        */
  /*                                                                             */
  /*                                                                             */

  /* OR if available use "licence.h"                                             */
  INIT_LICENCE();

  /*                                                                             */
  /*******************************************************************************/

  
  /////////////////////////////////////////////////////////////////////////////////
  // version information
  /////////////////////////////////////////////////////////////////////////////////

  if(1)
  {
    I32 Version, SubVersion;

    print("\nLinked with following library versions:\n");

    // GetLibVer_FASTLIB (&Version, &SubVersion);
    // print("FASTLIB : %d_%d %s\n", Version, SubVersion, GetLibVerString_FASTLIB());

    GetLibVer_VCLIB (&Version, &SubVersion);
    print("VCLIB   : %d_%d %s\n", Version, SubVersion, GetLibVerString_VCLIB());

    GetLibVer_VCRTLIB (&Version, &SubVersion);
    print("VCRTLIB : %d_%d %s\n", Version, SubVersion, GetLibVerString_VCRTLIB());

    print("\nCompiled with following header versions:\n");

    print("FLIB.H  : %u.%d\n", FASTLIB_VERSION, FASTLIB_SUBVERSION);
    print("VCLIB.H : %u.%d\n", VCLIB_VERSION, VCLIB_SUBVERSION);
    print("VCRT.H  : %u.%d\n", VCRTLIB_VERSION, VCRTLIB_SUBVERSION);

    print("\nVC/RT OS   running  with Version: %u.%d\n\n", getvar(VERSION), getvar(TESTVERSION));

    print("Compiled at %s %s Camera=%d\n\n", __DATE__, __TIME__, getvar(MODEL));
  }


  /////////////////////////////////////////////////////////////////////////////////
  // ini camera
  /////////////////////////////////////////////////////////////////////////////////

  // stop live image mode
  tpict();

  // set trigger output to exposer mode
  TRIGOUT_EXP();

  // set trigger input to positive edge
  TRIGINP_POS();

  // set logical page to standard image acquisition page
  ScrSetLogPage((I32)ScrGetCaptPage);  


  /////////////////////////////////////////////////////////////////////////////////
  // ini laser scanner parameter
  /////////////////////////////////////////////////////////////////////////////////

  // laser scanner parameter initialisation
  Error = InitSensorParameter(LsPar, MEM_INI);
  
  print("\n");


  /////////////////////////////////////////////////////////////////////////////////
  // ini scanner memory and lens calibration 
  /////////////////////////////////////////////////////////////////////////////////

  Error = ScannerInit(LsPar, MEM_INI);

  if(LsPar->CalibrationErrorKanatani == 0) print("Kanatani Calibration Ready.\n");
  if(LsPar->CalibrationErrorTsai     == 0) print("Tsai     Calibration Ready.\n");

  if(Error)
  {
    if(LsPar->CalibrationErrorKanatani)
    {
      print("Kanatani Initialisation Error %d: %s\n", LsPar->CalibrationErrorKanatani, LsPar->KanataniErrorMessage);
    }

    if(LsPar->CalibrationErrorTsai)
    {
      print("Tsai Initialisation Error: ");
      if(LsPar->CalibrationErrorTsai == ERR_FILE_OPEN) {print("Missing system calibration file.\n");}
      if(LsPar->CalibrationErrorTsai == ERR_FILE_SIZE) {print("Wrong system calibration file size\n");}
      if(LsPar->CalibrationErrorTsai == ERR_FILE_DATA) {print("Corrupted system calibration file\n");}
    }

    if(Error     == ERR_MEM_INI  ) {print("Error: Memory error in ScannerInit()\n"); goto QUIT;}
  }

  print("\n");

// no Kanatani available at test camera 7900154
// if(LsPar->CameraSNR == 7900154) {LsPar->CalibrationErrorKanatani = 1; print("Use Tsai for camera SNR %d\n", LsPar->CameraSNR);}


  /////////////////////////////////////////////////////////////////////////////////
  // ini product parameter
  /////////////////////////////////////////////////////////////////////////////////
  
  #ifdef PRODUCT_VERSION
  
    Error = IniProductParameter(LsPar, MEM_INI);
    
    if(Error == 0)
    {
      print("Reading Product Setting File %s successful.\n", PRODUCT_FILE);
    }
    else
    {
      if(Error == ERR_FILE_OPEN) {print("Error %d: No Product Setting File %s. Default parameter will be used.\n", Error, PRODUCT_FILE);}
      if(Error == ERR_FILE_SIZE) {print("Error %d: Missing some Product Parameter in File %s. Default parameter will be used.\n", Error, PRODUCT_FILE);}
      if(Error == ERR_FILE_DATA) {print("Error %d: Corrupted Product Parameter Data.\n", Error);}

      if(Error == ERR_MEM_INI  ) {print("Error: Memory error in IniProductParameter()\n"); goto QUIT;}
    }
  #endif


  /////////////////////////////////////////////////////////////////////////////////
  // allocate image buffer for parallel mode
  /////////////////////////////////////////////////////////////////////////////////

  Error = ScreenBuffer2(LsPar, SCREEN_BUFF_INI);
  if(Error == ERR_MEMORY) {print("Error %d: ScreenBuffer2()\n", Error); goto QUIT;}



  /////////////////////////////////////////////////////////////////////////////////
  // ini TCP/IP connection
  /////////////////////////////////////////////////////////////////////////////////

  response = TCP_Open(SERVER_PORT, &sock);

  if (response ==  0) print("\nNew stream socket created at port %d\n", SERVER_PORT);
  if (response == -1) print("\nCreate stream socket failed\n");
  if (response == -2) print("\nStream bind failed\n");
  if (response == -3) print("\nListen failed\n");

  if (response !=  0) goto QUIT;



  /////////////////////////////////////////////////////////////////////////////////
  // RUN
  /////////////////////////////////////////////////////////////////////////////////

  // test settings
  if(0)
  {
    LsPar->DebugDraw = 3;
    LsPar->DebugPrint = 2;
    // LsPar->SkipNoLaserPoints = 1;

    LsPar->NBR_LINES = -1;
    LsPar->MODE = -1; // No TCP/IP Data Transfer after camera boot 

    LsPar->SHUTTER = 200;
    LsPar->LASER_MODE = 1;
    
    // test ASCII
    // LsPar->CommandModeASCII = 1;
    // LsPar->DebugDraw = 3;
    // LsPar->PrintHostCommand = 7;
     
    // LsPar->TRIGGER = -1; 
    // LsPar->AutoTriggerFPS = 100;
    // LsPar->AutoTriggerError     =   0;
    // LsPar->EthernetPackNr = 2;
    // LsPar->ExposureMode = 1;
    // LsPar->ROI_DY = 256;
    // LsPar->ALGORITHM = 0;
    // LsPar->PrintHostCommand = 3;
    // LsPar->HorLine = 0;
    // LsPar->DoubleShutter[0] = 200;
    // LsPar->DoubleShutter[1] = 4000;
    // LsPar->SubSample = 2;
  }

  /////////////////////////////////////////////////////////////////////////////////
  // start 3D Scanner
  /////////////////////////////////////////////////////////////////////////////////

  print("\n--- ");
  print("%s Version: Main=%d / Lib=%d", "VC 3D Laser Scanner", LsPar->MainVersion, LsPar->LibVersion);
  #ifdef PRODUCT_VERSION
    print(" / Product=%d", ProdPar->ProductVersion);
  #endif
  print(" ---\n\n");

  print("Press ENTER for hand operated scan\n");
  print("Press  ESC  to quit program\n");
  print("Press  'm'  to change parameter\n");
  print("Press  's'  to save BMP image for debug\n");
  print("Press  'r'  to read BMP image for debug\n");
  print("\n");

  
  /////////////////////////////////////////////////////////////////////////////////
  // reset sensor and set scanner parameter
  /////////////////////////////////////////////////////////////////////////////////

  SensorReset(LsPar);


  /////////////////////////////////////////////////////////////////////////////////
  // clear output matrix
  /////////////////////////////////////////////////////////////////////////////////
  
  outPLC(0);


  /////////////////////////////////////////////////////////////////////////////////
  // scanning loop
  /////////////////////////////////////////////////////////////////////////////////

  do
  {
    //////////////////////////////////////////////
    // set start values
    //////////////////////////////////////////////

    if(getvar(ETHLINK) == 0)  LsPar->MODE = -1; // No TCP/IP Data Transfer, if disconnected


    if ( (LsPar->NBR_LINES == 0) || (LsPar->MODE == 0) )
    {
      // prepare sensor to wait for the first image
      ScreenBuffer2(LsPar, SCREEN_BUFF_FIRST);

      // reset auto trigger timer
      AutoTrigMode = 0;

      // reset double exposer counter
      LsPar->DoubleShutterImg = 0;

      // send remaining TCP/IP data
      TCP_SendCommand(LsPar, &alias, LsPar->SendBuffer, LsPar->MaxSendBuff, EMPTY_SEND_BUFFER);
    }


    //////////////////////////////////////////////
    // look for PLC, Ethernet or telnet input
    //////////////////////////////////////////////

    // telnet input
    response = ReadKeyboard_Telnet(0);
    if (response == KEY_ENTER) {if (LsPar->NBR_LINES == 0) LsPar->NBR_LINES++;} // single shot 
    if (response == KEY_ESC  ) {RunMode = 0;}
    if (response == 'm'      ) {ChangeParameter(LsPar); AutoTrigMode = 0;}
    if (response == 's'      ) {StoreImage();}
    if (response == 'r'      ) {ReadImage(); goto CALC_READ_IMAGE;}

    // Ethernet input
    response = TCP_ReadCommand(sock, &alias, (char *)LsPar->ReceiveBuffer, LsPar->MinReceiveBuff, LsPar->MaxReceiveBuff, &count);
    if (response == 0) response = TCP_RunCommand(LsPar, LsPar->ReceiveBuffer, count, &alias);


    //////////////////////////////////////////////
    // do image acquisition
    //////////////////////////////////////////////

    if( (LsPar->MODE) && (LsPar->NBR_LINES != 0) )
    {
      I32 ReadyImage  = SCREEN_RES_INI;

      // auto trigger active? (AUTO_TRIGGER_MODE [-1], AUTO_TRIGGER_POS after pos. edge [-2] or AUTO_TRIGGER_NEG after neg. edge [-3])
      if(LsPar->TRIGGER < 0)
      {
        static I32 TriggerStatus = -1;

        if(AutoTrigMode == 0)
        {
          ReadyImage = SCREEN_RES_BUSY;

          switch(LsPar->TRIGGER)
          {
            case  AUTO_TRIGGER_POS:
                  if(TRIGINP_PIN == 0                      ) TriggerStatus = 0;
                  if(TRIGINP_PIN != 0 && TriggerStatus == 0) TriggerStatus = 1;
                  break;

            case  AUTO_TRIGGER_NEG:
                  if(TRIGINP_PIN != 0                      ) TriggerStatus = 0;
                  if(TRIGINP_PIN == 0 && TriggerStatus == 0) TriggerStatus = 1;
                  break;

            default:
                  TriggerStatus = 1;
          }

          if(TriggerStatus == 1)
          {
            AutoTrigger(LsPar, AUTO_TRIG_INI);

            ReadyImage    = SCREEN_RES_INI;
            AutoTrigMode  =  1;
            TriggerStatus = -1;
          }
        }
        else
        {
          if( (LsPar->ExposureMode != DOUBLE_EXPOSURE) || (LsPar->DoubleShutterImg == 1) )
          {
            ReadyImage = SCREEN_RES_BUSY;

            if(AutoTrigger(LsPar, AUTO_TRIG_DIFF) == AUTO_TRIG_READY)
            {
              ReadyImage = ScreenBuffer2(LsPar, SCREEN_BUFF_SWAP);
              if(ReadyImage != SCREEN_RES_IMAGE) LsPar->AutoTriggerError++;
            }
          }
        }
      }


      // start next image acquisition if an image is ready
      if(ReadyImage == SCREEN_RES_INI)
      {
        ReadyImage = ScreenBuffer2(LsPar, SCREEN_BUFF_SWAP);
      }


      //////////////////////////////////////////////
      // calc image
      //////////////////////////////////////////////
      
      if (ReadyImage == SCREEN_RES_IMAGE)
      {
        CALC_READ_IMAGE:
      
        //////////////////////////////////////////////
        // detect laser line
        //////////////////////////////////////////////

        Ms1 = 1000 * getlvar(SEC) + getvar(MSEC);
        {
          image ImgArea;
          
          ImageAssign(&ImgArea, ScrByteAddr(0, 0), LsPar->ROI_DX, LsPar->ROI_DY_SubSample, ScrGetPitch);
          
          Error = ScannerDetection(LsPar, &ImgArea);
        }
        MsD = 1000 * getlvar(SEC) + getvar(MSEC) - Ms1;

        if(Error) print("ScannerDetection Error %d.\n", Error);
        

        //////////////////////////////////////////////
        // calc auto shutter value
        //////////////////////////////////////////////

        if(LsPar->ExposureMode == AUTO_EXPOSURE) AutoShutter(LsPar);
          
          
        //////////////////////////////////////////////
        // product calculation
        //////////////////////////////////////////////

        if(ReadyForCalc(LsPar))
        {
          // select all valid detections (if not, check if detection is inside ROI or ROI_MM)
          if(LsPar->SkipNoLaserPoints) DelNoneDetectionPoints(LsPar);
          
          // compensate sub sample
          if(LsPar->SubSample) 
          {
            I32 i, Fact, Nr = LsPar->LaserPoints;
            
            Fact = 1 << LsPar->SubSample;
            for(i=0; i<Nr; i++) LsPar->pLaserPosZ[i] *= Fact;
          }
          
          // convert from IMAGE to WORLD coordinates for measurements (mm)
          ImgToWorldCoord(LsPar, LsPar->LaserPoints, LsPar->pLaserPosX, LsPar->pLaserPosZ, LsPar->pLaserPosX_MM, LsPar->pLaserPosZ_MM);
          
          // calc product results
          #ifdef PRODUCT_VERSION
          Product_Measure(LsPar);
          #endif

          if(LsPar->NBR_LINES > 0) LsPar->NBR_LINES--;

          //////////////////////////////////////////////
          // send results via TCP/IP
          //////////////////////////////////////////////
          
          TCP_SendCommand(LsPar, &alias, LsPar->SendBuffer, LsPar->MaxSendBuff, USE_SEND_BUFFER);
        }

        if(LsPar->DebugPrint == 1) print("SysMemFree=%6d\n", sysmemfree());
        if(LsPar->DebugPrint == 2) {static I32 cnt=0; if(++cnt == 50) {print("t=%2dms fps=%5.1f  Error=%d\n", (I32)MsD, (float)(cnt*1000)/(float)(Ms1-Ms0), Error); Ms0=Ms1; cnt=0;}}
      }
    }
  }
  while(RunMode);

  /////////////////////////////////////////////////////////////////////////////////
  // Stop RUN mode
  /////////////////////////////////////////////////////////////////////////////////

  QUIT:

  if(alias) TCP_Close(alias);
  if(sock)  TCP_Close(sock);

  ScreenBuffer2(LsPar, SCREEN_BUFF_DEINI);

  #ifdef PRODUCT_VERSION
    IniProductParameter(LsPar, MEM_DEINI);
  #endif

  InitSensorParameter(LsPar, MEM_DEINI);
  ScannerInit(LsPar, MEM_DEINI);

  vmode(vmLive);
  

  /////////////////////////////////////////////////////////////////////////////////
  // set lower task priority
  /////////////////////////////////////////////////////////////////////////////////

  task_set_priority(id, oldprio, &highprio); // reset Taskpriority (old value)
  print("Set own task '%s' to prio %d! (highprio=%d)\n", task_get_name_from_id(id), oldprio, highprio);
}

/*************************************************************************************************************************/

I32 InitSensorParameter (LsParameter *LsPar, I32 Mode)
{
  I32        Debug=0, Error=0;
  static I32 MemoryReady=0;
  

  /////////////////////////////////////////////////////////////////////////////////
  // release memory
  /////////////////////////////////////////////////////////////////////////////////

  if(Mode == MEM_DEINI || Mode == MEM_INI)
  {
    if(MemoryReady) 
    {
      MemoryReady = 0;
      
      if(LsPar->SendBuffer     != NULL) sysfree(LsPar->SendBuffer);
      if(LsPar->ReceiveBuffer  != NULL) sysfree(LsPar->ReceiveBuffer);
      if(LsPar->pSensorSetting != NULL) sysfree(LsPar->pSensorSetting);
    }
  }


  /////////////////////////////////////////////////////////////////////////////////
  // ini parameter and allocate memory
  /////////////////////////////////////////////////////////////////////////////////

  if( Mode == MEM_INI || (MemoryReady == 0 && Mode != MEM_DEINI) )
  {
    Error       = 0;
    MemoryReady = 1;
    
    // for debugging print host commands and drawings
    LsPar->PrintHostCommand  = 0;
    LsPar->DebugPrint        = 0;
    LsPar->DebugDraw         = 0;

    LsPar->pSensorSetting   = NULL;
    LsPar->ReceiveBuffer    = NULL;
    LsPar->SendBuffer       = NULL;

    // memory for sensor setting parameter
    LsPar->pSensorSetting    = (I8 *) sysmalloc( (MAX_SETTING_SIZE * sizeof(  U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->pSensorSetting == NULL) {Error = ERR_MEM_INI;}  

    // memory for receive buffer
    LsPar->MaxReceiveBuff    = CMD_HEAD_SIZE + MAX_SETTING_SIZE;
    LsPar->ReceiveBuffer     = (U8  *) sysmalloc( ((LsPar->MaxReceiveBuff) * sizeof(   U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->ReceiveBuffer  == NULL) {Error = ERR_MEM_INI;}

    // memory for send buffer [(ScrGetColumns * ScrGetRows + CMD_HEAD_SIZE) .. (sysmemfree() - 8 * ScrGetColumns * ScrGetRows)]
    LsPar->MaxSendBuff       = CMD_HEAD_SIZE + 4 * ScrGetColumns * ScrGetRows; // memory for 512 scans
    LsPar->SendBuffer        = (U8 *) sysmalloc( (LsPar->MaxSendBuff * sizeof(U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->SendBuffer     == NULL) {Error = ERR_MEM_INI;}
    
    if(Error) 
    {
      print("Error: Memory error in InitSensorParameter()\n");
      goto QUIT_INI;
    }
  }


  /////////////////////////////////////////////////////////////////////////////////
  // read sensor parameter file
  /////////////////////////////////////////////////////////////////////////////////

  if(Mode == MEM_INI || Mode == READ_DATA)
  {
    I32 Result;
  
    Error  = 0;
    Result = 0;
    
    // debug prints
    if(LsPar->DebugPrint == 8) Debug = 1; else Debug = 0;
    // Debug = 1;

    // copy data file from flash to RAM
    Result = ReadFileToMem(SENSOR_FILE, LsPar->pSensorSetting, MAX_SETTING_SIZE, Debug);    

    if(Result > 0)
    {
      LsPar->SensorSettingSize  = Result;
    }
    else
    {
      Error = Result;
      
      LsPar->SensorSettingSize = 0;
      LsPar->pSensorSetting[0] = 0;
      
      print("Sensor parameter file %s doesn't exist. Error %d. \n", SENSOR_FILE, Error);
    }
  }
  

  /////////////////////////////////////////////////////////////////////////////////
  // set sensor parameter
  /////////////////////////////////////////////////////////////////////////////////

  if(Mode == MEM_INI || Mode == READ_DATA || Mode == SET_DATA)
  {
    // read parameter from RAM
    Error  = 0;
    Error += ReadIntParameter(LsPar, "LASER_MODE", &LsPar->LASER_MODE, Debug);
    Error += ReadIntParameter(LsPar, "ROI_X", &LsPar->ROI_X, Debug);
    Error += ReadIntParameter(LsPar, "ROI_Y", &LsPar->ROI_Y, Debug);
    Error += ReadIntParameter(LsPar, "ROI_DX", &LsPar->ROI_DX, Debug);
    Error += ReadIntParameter(LsPar, "ROI_DY", &LsPar->ROI_DY, Debug);
    Error += ReadIntParameter(LsPar, "GAINVAL", &LsPar->GAINVAL, Debug);
    Error += ReadIntParameter(LsPar, "SHUTTER", &LsPar->SHUTTER, Debug);
    Error += ReadIntParameter(LsPar, "TRIGGER", &LsPar->TRIGGER, Debug);
    Error += ReadIntParameter(LsPar, "MODE", &LsPar->MODE, Debug);
    Error += ReadIntParameter(LsPar, "NBR_LINES", &LsPar->NBR_LINES, Debug);
    Error += ReadIntParameter(LsPar, "METHOD", &LsPar->METHOD, Debug);
    Error += ReadIntParameter(LsPar, "ALGORITHM", &LsPar->ALGORITHM, Debug);
    Error += ReadIntParameter(LsPar, "RlcThresh", &LsPar->RlcThresh, Debug);
    Error += ReadIntParameter(LsPar, "MinWidth", &LsPar->MinWidth, Debug);
    Error += ReadIntParameter(LsPar, "MaxWidth", &LsPar->MaxWidth, Debug);
    Error += ReadIntParameter(LsPar, "ExposureMode", &LsPar->ExposureMode, Debug);
    Error += ReadIntParameter(LsPar, "AutoShutterVal", &LsPar->AutoShutterVal, Debug);
    Error += ReadIntParameter(LsPar, "AutoShutterMin", &LsPar->AutoShutterMin, Debug);
    Error += ReadIntParameter(LsPar, "AutoShutterMax", &LsPar->AutoShutterMax, Debug);
    Error += ReadIntParameter(LsPar, "DoubleShutter0", &LsPar->DoubleShutter[0], Debug);
    Error += ReadIntParameter(LsPar, "DoubleShutter1", &LsPar->DoubleShutter[1], Debug);
    Error += ReadIntParameter(LsPar, "ReflexionFilter", &LsPar->ReflexionFilter, Debug);
    Error += ReadIntParameter(LsPar, "EthernetPackNr", &LsPar->EthernetPackNr, Debug);
    Error += ReadIntParameter(LsPar, "AutoTriggerFPS", &LsPar->AutoTriggerFPS, Debug);
    Error += ReadIntParameter(LsPar, "AutoTriggerError", &LsPar->AutoTriggerError, Debug);
    Error += ReadIntParameter(LsPar, "EthernetSendNoWait", &LsPar->EthernetSendNoWait, Debug);
    Error += ReadIntParameter(LsPar, "SubSample", &LsPar->SubSample, Debug);
    Error += ReadIntParameter(LsPar, "MedianFilter", &LsPar->MedianFilter, Debug);
    Error += ReadIntParameter(LsPar, "SpeckleFilter", &LsPar->SpeckleFilter, Debug);
    Error += ReadIntParameter(LsPar, "SkipNoLaserPoints", &LsPar->SkipNoLaserPoints, Debug);
    Error += ReadIntParameter(LsPar, "CommandModeASCII", &LsPar->CommandModeASCII, Debug);

    if(Error) print("Missing %d sensor parameter setting(s). Default parameters will be used.\n", Error);
    else      print("Reading sensor parameter file %s successful.\n", SENSOR_FILE);
    
    // set ROI in mm
    ImgToWorld_ROI(LsPar);

    // Licence code for all VC3D Laser Scanner Cameras. If you want to use the lib on other VC cameras, please ask VC for a new licence code
    LsPar->LicenceCode1      = 0x674F27A1;
    LsPar->LicenceCode2      = 0x2E05F267;

    // set scanner default parameter
    LsPar->StepDx            = 1;

    // double exposure
    LsPar->DoubleShutterImg  = 0;

    // program version
    LsPar->MainVersion       = MAIN_VERSION;

    // read camera SNR
    LsPar->CameraSNR         = GetCameraID();
  
    // set scanner default parameter
    LsPar->STOP              = 0;  

    // laser detection parameter
    LsPar->StepDx            = 1;  
  }


  /////////////////////////////////////////////////////////////////////////////////
  // parameters depending on setting values
  /////////////////////////////////////////////////////////////////////////////////  
  
  {
    // CmdID + HostCounter + Action + Val = 4 INT values
    LsPar->MinReceiveBuff       = 4 * sizeof(I32); 
    
    // Binary or ASCII transfer (4 times 8 ASCII characters)
    if(LsPar->CommandModeASCII) LsPar->MinReceiveBuff = 4 * 8; 
  }

  /////////////////////////////////////////////////////////////////////////////////
  // ready
  /////////////////////////////////////////////////////////////////////////////////  

  QUIT_INI:
  
  LsPar->Error = Error;
  
  return Error;
}

/*************************************************************************************************************************/

I32 GetParameterRange (LsParameter *LsPar, void *pPar, I32 *SetVal, I32 *MinVal, I32 *MaxVal)
{
  I32 Error = 0;

  /////////////////////////////////////////////////////////////////////////////////
  // ini values
  /////////////////////////////////////////////////////////////////////////////////

  *MinVal = 0;
  *MaxVal = 0;
  *SetVal = 0;
  

  /////////////////////////////////////////////////////////////////////////////////
  // parameer range
  /////////////////////////////////////////////////////////////////////////////////

  if (pPar == &LsPar->STOP) 
  {
    *MinVal = 0;
    *MaxVal = 0;
    *SetVal = 0;
  }

  if (pPar == &LsPar->LASER_MODE) 
  {
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_X) 
  {
    *MinVal = 0;
    *MaxVal = ScrGetColumns - 3;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_Y) 
  {
    *MinVal = 0;
    *MaxVal = ScrGetRows - 3;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_DX)
  {
    *MinVal = 6;
    *MaxVal = ScrGetColumns;
    *SetVal = ScrGetColumns;
  }

  if (pPar == &LsPar->ROI_DY) 
  {
    *MinVal = 6;
    *MaxVal = ScrGetRows;
    *SetVal = ScrGetRows;
  }

  if (pPar == &LsPar->GAINVAL)
  {
    *MinVal = 0;
    *MaxVal = 1023;
    *SetVal = getvar(GAIN);
  }

  if (pPar == &LsPar->SHUTTER) 
  {
    *MinVal = 5;
    *MaxVal = 500000;
    *SetVal = 500;
  }

  if (pPar == &LsPar->TRIGGER) 
  {
    *MinVal = -3;
    *MaxVal =  2;
    *SetVal =  0;
  }

  if (pPar == &LsPar->MODE) 
  {
    *MinVal = -1;
    *MaxVal =  9;
    *SetVal =  0;
  }

  if (pPar == &LsPar->NBR_LINES) 
  {
    *MinVal = -1;
    *MaxVal = 10000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->METHOD) 
  { 
    *MinVal = 0;
    *MaxVal = 2;
    *SetVal = 2;
  }

  if (pPar == &LsPar->ALGORITHM) 
  {
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 0;
  }

  if (pPar == &LsPar->RlcThresh) 
  {
    *MinVal =   1;
    *MaxVal = 254;
    *SetVal =  48;
  }

  if (pPar == &LsPar->MinWidth)
  {
    *MinVal =   1;
    *MaxVal = 100;
    *SetVal =   1;
  }

  if (pPar == &LsPar->MaxWidth)
  {
    *MinVal =   1;
    *MaxVal = 100;
    *SetVal =  30;
  }

  if (pPar == &LsPar->ExposureMode) 
  {
    *MinVal = 0;
    *MaxVal = 3;
    *SetVal = FIX_EXPOSURE;
  }

  if (pPar == &LsPar->AutoShutterVal) 
  { 
    *MinVal = 1;
    *MaxVal = 254;
    *SetVal = 128;
  }

  if (pPar == &LsPar->AutoShutterMin) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 10;
  }

  if (pPar == &LsPar->AutoShutterMax) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 1000;
  }

  if (pPar == &LsPar->DoubleShutter[0]) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 100;
  }

  if (pPar == &LsPar->DoubleShutter[1])
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 500;
  }

  if (pPar == &LsPar->ReflexionFilter) 
  {
    *MinVal = -8;
    *MaxVal =  8;
    *SetVal =  0;
  }

  if (pPar == &LsPar->ROI_X_MM) 
  {
    *MinVal = -10000;
    *MaxVal = +10000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_Y_MM) 
  {
    *MinVal = 0;
    *MaxVal = 10000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_DX_MM) 
  {
    *MinVal = 0;
    *MaxVal = 100000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_DY_MM) 
  {
    *MinVal = 0;
    *MaxVal = 100000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->LibVersion) 
  { 
    *MinVal = 0;
    *MaxVal = 100;
    *SetVal = VERSION;
  }

  if (pPar == &LsPar->EthernetPackNr) 
  {
    *MinVal = 0;
    *MaxVal = 100;
    *SetVal = 1;
  }

  if (pPar == &LsPar->AutoTriggerFPS) 
  { 
    *MinVal = 1;
    *MaxVal = 10000;
    *SetVal = 50 * 10; // 1/10 Hz
  }

  if (pPar == &LsPar->AutoTriggerError) 
  { 
    *MinVal = 0;
    *MaxVal = 2147483647; // Max I32
    *SetVal = 0;
  }

  if (pPar == &LsPar->EthernetSendNoWait) 
  {
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 1;
  }

  if (pPar == &LsPar->SubSample)
  {
    *MinVal = 0;
    *MaxVal = 2;
    *SetVal = 0;
  }

  if (pPar == &LsPar->CameraSNR) 
  { 
    *MinVal = 1000000;
    *MaxVal = 9999999;
    *SetVal = GetCameraID();
;
  }

  if (pPar == &LsPar->MedianFilter) 
  { 
    *MinVal = 0;
    *MaxVal = 4;
    *SetVal = 0;
  }

  if (pPar == &LsPar->SpeckleFilter) 
  { 
    *MinVal = 0;
    *MaxVal = 4;
    *SetVal = 3;
  }

  if (pPar == &LsPar->SkipNoLaserPoints) 
  { 
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 1;
  }

  if (pPar == &LsPar->InPLC) 
  { 
    *MinVal = 0;
    *MaxVal = 15;
    *SetVal = 0;
  }

  if (pPar == &LsPar->OutPLC) 
  { 
    *MinVal = 0;
    *MaxVal = 15;
    *SetVal = 0;
  }
  
  if (pPar == &LsPar->CommandModeASCII) 
  { 
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 0;
  }
  
  /////////////////////////////////////////////////////////////////////////////////
  // check value
  /////////////////////////////////////////////////////////////////////////////////

  if(*SetVal < *MinVal) *SetVal = *MinVal;
  if(*SetVal > *MaxVal) *SetVal = *MaxVal;  
  
  
  return Error;
}

/*************************************************************************************************************************/

I32 ReadIntParameter (LsParameter *LsPar, I8 *pParFile, I32 *pParStruct, I32 Debug)
{
  I32 SetVal, MinVal, MaxVal, Error;
  
  GetParameterRange(LsPar, pParStruct, &SetVal, &MinVal, &MaxVal);
  
  Error = ReadIntVar(LsPar->pSensorSetting, pParFile, SetVal, MinVal, MaxVal, pParStruct, Debug);
  
  if(Error) Error = 1;
  
  return Error;
}

/*************************************************************************************************************************/

I32 StoreSensorParameter (LsParameter *LsPar)
{
  I32    debug=0, write=1, Error;
  FILE   *fp;  

  print("Store file %s. Please wait...\n", SENSOR_FILE);

  // open write file
  fp = OpenFile(SENSOR_FILE, debug, write);
  
  if (fp)
  { 
    Error  = 0;
    
    Error += WriteIntToFile(fp, "LASER_MODE", LsPar->LASER_MODE);
    Error += WriteIntToFile(fp, "ROI_X", LsPar->ROI_X);
    Error += WriteIntToFile(fp, "ROI_Y", LsPar->ROI_Y);
    Error += WriteIntToFile(fp, "ROI_DX", LsPar->ROI_DX);
    Error += WriteIntToFile(fp, "ROI_DY", LsPar->ROI_DY);
    Error += WriteIntToFile(fp, "GAINVAL", LsPar->GAINVAL);
    Error += WriteIntToFile(fp, "SHUTTER", LsPar->SHUTTER);
    Error += WriteIntToFile(fp, "TRIGGER", LsPar->TRIGGER);
    Error += WriteIntToFile(fp, "MODE", LsPar->MODE);
    Error += WriteIntToFile(fp, "NBR_LINES", LsPar->NBR_LINES);
    Error += WriteIntToFile(fp, "METHOD", LsPar->METHOD);
    Error += WriteIntToFile(fp, "ALGORITHM", LsPar->ALGORITHM);
    Error += WriteIntToFile(fp, "RlcThresh", LsPar->RlcThresh);
    Error += WriteIntToFile(fp, "MinWidth", LsPar->MinWidth);
    Error += WriteIntToFile(fp, "MaxWidth", LsPar->MaxWidth);
    Error += WriteIntToFile(fp, "ExposureMode", LsPar->ExposureMode);
    Error += WriteIntToFile(fp, "AutoShutterVal", LsPar->AutoShutterVal);
    Error += WriteIntToFile(fp, "AutoShutterMin", LsPar->AutoShutterMin);
    Error += WriteIntToFile(fp, "AutoShutterMax", LsPar->AutoShutterMax);
    Error += WriteIntToFile(fp, "DoubleShutter0", LsPar->DoubleShutter[0]);
    Error += WriteIntToFile(fp, "DoubleShutter1", LsPar->DoubleShutter[1]);
    Error += WriteIntToFile(fp, "ReflexionFilter", LsPar->ReflexionFilter);
    Error += WriteIntToFile(fp, "EthernetPackNr", LsPar->EthernetPackNr);
    Error += WriteIntToFile(fp, "AutoTriggerFPS", LsPar->AutoTriggerFPS);
    Error += WriteIntToFile(fp, "AutoTriggerError", LsPar->AutoTriggerError);
    Error += WriteIntToFile(fp, "EthernetSendNoWait", LsPar->EthernetSendNoWait);
    Error += WriteIntToFile(fp, "SubSample", LsPar->SubSample);
    Error += WriteIntToFile(fp, "MedianFilter", LsPar->MedianFilter);
    Error += WriteIntToFile(fp, "SpeckleFilter", LsPar->SpeckleFilter);
    Error += WriteIntToFile(fp, "SkipNoLaserPoints", LsPar->SkipNoLaserPoints);
    Error += WriteIntToFile(fp, "CommandModeASCII", LsPar->CommandModeASCII);
    
    if(Error) print("Can't write file %s\n", SENSOR_FILE);
  }
  else
  {
    Error = ERR_FILE_OPEN;
    print("Can't open file %s for write\n", SENSOR_FILE);
  }
  
  // close file 
  CloseFile(fp);
  
  return Error;
}

/*************************************************************************************************************************/

void ChangeParameter (LsParameter *LsPar)

{
  I32 response;

  EmptyKeyboardBuffer();

  do
  {
    /********************************************/
    /* open menu                                */
    /********************************************/

    print("\n");
    print("--------------------------------------------------------------------------------\n");
    print("Setup Parameter\n");
    print("--------------------------------------------------------------------------------\n");
    print("l:   live-picture\n");
    print("p:   set standard parameter\n");
    print("e:   set expert parameter\n");
    print("s:   save sensor parameter\n");
    print("r:   read sensor parameter\n");
    print("d:   debug parameters\n");
    #ifdef PRODUCT_VERSION
    print("t:   read product parameter\n");
    #endif
    print("--------------------------------------------------------------------------------\n");
    print("q:   quit Menu\n");
    print("\n");


    response = ReadKeyboard_Telnet(1);

    /********************************************/
    /* menu                                     */
    /********************************************/

    switch(response)
    {
      case 'l' :
                 vmode(vmLive);
                 break;

      case 'p' :
                 ChangeValue_Sensor(LsPar, &LsPar->STOP,               1, "Reset image acquisition");
                 ChangeValue_Sensor(LsPar, &LsPar->LASER_MODE,         1, "Set Laser: 0=OFF 1=ON");
                 ChangeValue_Sensor(LsPar, &LsPar->TRIGGER,            1, "Set Trigger: 0=OFF 1=PosEdge 2=NegEdge -1=AutoTrigger -2=AutoTrigPos -3=AutoTrigNeg");
                 if(LsPar->TRIGGER < 0)
                 {
                 ChangeValue_Sensor(LsPar, &LsPar->AutoTriggerFPS,    10, "Set AutoTrigger Frames Per Second [1/10 Hz]");
                 ChangeValue_Sensor(LsPar, &LsPar->AutoTriggerError,   1, "Set AutoTrigger Error");
                 }
                 ChangeValue_Sensor(LsPar, &LsPar->ROI_X,              8, "Set ROI Area X [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->ROI_Y,              8, "Set ROI Area Y [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->ROI_DX,             8, "Set ROI Area DX [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->ROI_DY,             8, "Set ROI Area DY [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->NBR_LINES,          1, "Set Nr_Lines: -1=endless X=NrImages");
                 ChangeValue_Sensor(LsPar, &LsPar->MODE,               1, "Set Data Mode");
                 ChangeValue_Sensor(LsPar, &LsPar->ExposureMode,       1, "Set Exposure Mode: 0=FIX_EXPOSURE 1=AUTO_EXPOSURE 2=DOUBLE_EXPOSURE");

                 if(LsPar->ExposureMode == FIX_EXPOSURE)
                 {
                 ChangeValue_Sensor(LsPar, &LsPar->SHUTTER,            5, "Set Shutter [us]");
                 }

                 if(LsPar->ExposureMode == AUTO_EXPOSURE)
                 {
                 ChangeValue_Sensor(LsPar, &LsPar->AutoShutterVal,     1, "Auto shutter Laser brightness");
                 ChangeValue_Sensor(LsPar, &LsPar->AutoShutterMin,     5, "Auto shutter time min [us]");
                 ChangeValue_Sensor(LsPar, &LsPar->AutoShutterMax,    50, "Auto shutter time max [us]");
                 }

                 if(LsPar->ExposureMode == DOUBLE_EXPOSURE)
                 {
                 ChangeValue_Sensor(LsPar, &LsPar->DoubleShutter[0],   1, "Double Exposure (Shutter 1 [us])");
                 ChangeValue_Sensor(LsPar, &LsPar->DoubleShutter[1],   1, "Double Exposure (Shutter 2 [us])");
                 }
                 break;

      case 'e' :
                 ChangeValue_Sensor(LsPar, &LsPar->GAINVAL,            1, "Set GAINVAL");
                 ChangeValue_Sensor(LsPar, &LsPar->METHOD,             1, "Set Methode");
                 ChangeValue_Sensor(LsPar, &LsPar->ALGORITHM,          1, "Set Algorithm (0=I16 / 1=F32)");
                 ChangeValue_Sensor(LsPar, &LsPar->RlcThresh,          1, "RLC Threshold");
                 ChangeValue_Sensor(LsPar, &LsPar->SubSample,          1, "Image Sub Sample");
                 ChangeValue_Sensor(LsPar, &LsPar->StepDx,             1, "Step Dx [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->MinWidth,           1, "Min. Laser Width [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->MaxWidth,           1, "Max. Laser Width [Pixel]");
                 ChangeValue_Sensor(LsPar, &LsPar->EthernetPackNr,     1, "Ethernet Line Package Nr");
                 ChangeValue_Sensor(LsPar, &LsPar->EthernetSendNoWait, 1, "Ethernet Send No Wait Mode");
                 ChangeValue_Sensor(LsPar, &LsPar->ReflexionFilter,    1, "Laser Detection Filter");
                 ChangeValue_Sensor(LsPar, &LsPar->MedianFilter,       1, "Median Filter: 0=no filter 1=3x1 2=5x1 3=7x1 4=9x1");
                 ChangeValue_Sensor(LsPar, &LsPar->SpeckleFilter,      1, "Speckle Filter: 0=no filter 1=3x3 2=5x5 3=7x7 4=9x9");
                 ChangeValue_Sensor(LsPar, &LsPar->SkipNoLaserPoints,  1, "Skip no laser detections points");
                 ChangeValue_Sensor(LsPar, &LsPar->CommandModeASCII,   1, "Command transfer mode (0=Binary / 1=ASCII)");
                 break;
                 
      case 'd' :
                 ChangeValue_Telnet(&LsPar->DebugPrint,       0,  8,  1, "Debug Print: 1=Memory 2=fps 3=Shutter 4=ImgAdr 5=TCP/IP 6=ImgToWorld 7=ProdRes 8=ParSet");
                 ChangeValue_Telnet(&LsPar->DebugDraw,        0,  3,  1, "Debug Draw: 1=LaserCenter 2=LaserDetection 3=ProductResults");
                 ChangeValue_Telnet(&LsPar->PrintHostCommand, 0,  7,  1, "Debug Commands: Bit0=Host->Cam Bit1=Cam->Host Bit2=Data");
                 break;
                 
      case 'r' :
                 InitSensorParameter(LsPar, READ_DATA);
                 break;
                 
      case 's' :
                 StoreSensorParameter(LsPar);
                 break;
                 
      #ifdef PRODUCT_VERSION
      case 't' :
                 {
                   I32 Error;
                   
                   print("Reading Product Setting File %s ...\n", PRODUCT_FILE);
                   Error = IniProductParameter(LsPar, READ_DATA);
                   if(Error == 0)
                   {
                     print("Ready.\n");
                   }
                   else
                   {
                     if(Error == ERR_FILE_OPEN) {print("Error %d: No Product Setting File %s. Default parameter will be used.\n", Error, PRODUCT_FILE);}
                     if(Error == ERR_FILE_SIZE) {print("Error %d: Missing some Product Parameter in File %s. Default parameter will be used.\n", Error, PRODUCT_FILE);}
                     if(Error == ERR_FILE_DATA) {print("Error %d: Corrupted Product Parameter Data.\n", Error);}
                     if(Error == ERR_MEM_INI  ) {print("Error: Memory error in IniProductParameter()\n");}
                   }
                 }
                 break; 
      #endif
    }
  }
  while(response != 'q');


  /////////////////////////////////////////////
  // reset sensor with new settings
  /////////////////////////////////////////////

  SensorReset(LsPar);

  print("RUN mode\n");
}

/*************************************************************************************************************************/

void ChangeValue_Sensor (LsParameter *LsPar, I32 *ParValue, I32 Step, I8 *text)
{
  I32 SetVal, MinVal, MaxVal;
  
  // get range
  GetParameterRange(LsPar, ParValue, &SetVal, &MinVal, &MaxVal);
  
  // change value
  ChangeValue_Telnet(ParValue, MinVal, MaxVal, Step, text);
}

/*************************************************************************************************************************/

void ChangeValue_Telnet (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text)
{
  I32  value;
  I32  response=0; 

  // start value
  value = *ParValue;

  // change value
  do
  {
    if(response == KEY_RIGHT) {value+=Step; if(value > Max) value=Max;}
    if(response == KEY_LEFT ) {value-=Step; if(value < Min) value=Min;}

    print("%s = %d  [%d - %d]   \r", text, value, Min, Max);

    response=ReadKeyboard_Telnet(1);
  }
  while( (response != KEY_ESC) && (response != KEY_ENTER) );

  // take new value
  if (response == KEY_ENTER)  *ParValue = value;

  print("%s = %d  [%d - %d]   \n", text, value, Min, Max);
}

/*************************************************************************************************************************/

#define TimeOut 10 /* ms */

int ReadKeyboard_Telnet (int wait)
{
  int i, ch1, ch2, ch3, ch4, src;


  if(wait || rbready())
  {

  while(1)
  {
    // if (kbready()) {src=1; break;}
    if (rbready()) {src=0; break;}
    time_delay(1);
  }

  if(src) ch1 = kbdrcv();
  else    ch1 = rs232rcv();

  if (ch1 != 0x1b)
     if(ch1 == 0x0d || ch1 == 0x0a)
       return(KEY_ENTER);
     else
       return(ch1);

  ch1 = KEY_ESC;

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if ( kbhit() ) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch2 = kbdrcv();
  else    ch2 = rs232rcv();

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if ( kbhit() ) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch3 = kbdrcv();
  else    ch3 = rs232rcv();


  if (ch2 == 0x4f)   /* 2/3 of a F1 or F2 received? */
  {

    if ((ch3 == 0x71) || (ch3 == 0x50))   /* F1 received? */
       return(KEY_F1);

    if ((ch3 == 0x72) || (ch3 == 0x51))   /* F2 received? */
       return(KEY_F2);

    return(KEY_INVALID);
  }

  if (ch2 == 0x5B)   /* 2/3 of a Cursor received? */
  {
    if (ch3 == 0x31)
    {
      for (i=0; i<TimeOut; i++)
      {
        time_delay(1);
        if(src) {if (kbready()) break;}
        else    {if ( kbhit() ) break;}
      }
      if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

      if(src) ch4 = kbdrcv();
      else    ch4 = rs232rcv();

      if (ch4 == 0x31)   /* F1 received? */
       return(KEY_F1);

      if (ch4 == 0x32)   /* F2 received? */
       return(KEY_F2);
    }
    else
    {
      switch (ch3)
    {
      case 0x41 :
      case 0x42 :
      case 0x43 :
      case 0x44 : return(ch3 - 61);
      default   : return(KEY_INVALID);
    }
     }
  }
  return(KEY_INVALID);
  }
  else
  {
  return(KEY_INVALID);
  }
}

/************************************************************************************************************************/

I32 ReadyForCalc (LsParameter *LsPar)
{
  if(LsPar->ExposureMode == DOUBLE_EXPOSURE)
  {
    if (LsPar->DoubleShutterImg)
    {
      LsPar->DoubleShutterImg = 0;

      return 1;
    }
    else
    {
      LsPar->DoubleShutterImg = 1;

      return 0;
    }
  }
  else
  {
    return 1;
  }
}

/***************************************************************************************************************************/

I32 ScreenBuffer2 (LsParameter *LsPar, I32 mode)
{
  I32        NrOfScreens = 2, Trigger = 0;
  static I32 Screen[2], TrackNr[2], ScrAdr=0, TakeFirstImage = 0, TakeNextImage = 0, calc, capt;

  // print("ScreenBuffer2:  Mode=%d\n", mode);

  if(LsPar->TRIGGER > 0) Trigger = 1;

  /*
  #define SCREEN_BUFF_INI        0
  #define SCREEN_BUFF_SWAP       1
  #define SCREEN_BUFF_RESET      2
  #define SCREEN_BUFF_DEINI      3
  #define SCREEN_BUFF_FIRST      4
  */

  // screen allocation mode
  if (mode == SCREEN_BUFF_INI)
  {
    Screen [0] = ScrGetCaptPage;

    ScrAdr     = (I32)DRAMScreenMalloc();
    if(ScrAdr == 0) return ERR_MEMORY;
    Screen [1] = (ScrAdr & 0xFFFFFC00) + 1024;

    // now reset
    mode = SCREEN_BUFF_RESET;
  }


  // reset sensor (cancel images, new format etc.)
  if (mode == SCREEN_BUFF_RESET)
  {
    TrackNr[0] = getvar(IMGREADY) + 1;
    TrackNr[1] = getvar(IMGREADY) + 1;

    while(cancel_capture_rq());
    SetImageFormat(LsPar);

    // after SetImageFormat, the sensor needs a dummy image, as the first image still has the old format
    tpict();

    capt = 0;
    calc = 1;

    // print("RESET: TrackNr[%d]=%d\n", calc, TrackNr[calc]);

    // now ini variables
    mode = SCREEN_BUFF_FIRST;
  }


  // prepare sensor for first image
  if (mode == SCREEN_BUFF_FIRST)
  {
    TakeFirstImage = 1;
  }


  // change screen for calculation and acquisition
  if (mode == SCREEN_BUFF_SWAP)
  {
    I32 DeltaImg, Sh;

    // print("SWAP:  TrackNr[%d]=%d  IMGREADY=%d  Screen[%d]=%d\n", calc, TrackNr[calc], getvar(IMGREADY), capt, Screen[capt]);

    // take start image (fill image pipe)
    if(TakeFirstImage)
    {
      TakeFirstImage = 0;

      if( LsPar->ExposureMode == DOUBLE_EXPOSURE) Sh = LsPar->DoubleShutter[capt];
      else                                        Sh = LsPar->SHUTTER;

      shutter(Sh);

      TrackNr[capt] = capture_request(getvar(EXPCNT), getvar(GAIN), (I32 *)Screen[capt], Trigger);

      if(LsPar->DebugPrint == 4) print("Capt. Img 1: t=%3dms  CapPage=0x%X  Sh=%4d  Trig=%d  Track[%d]=%3d  DoubleShutterImg=%d  NBR_LINES=%d\n", getvar(MSEC), Screen[capt], Sh, Trigger, capt, TrackNr[capt], LsPar->DoubleShutterImg, LsPar->NBR_LINES);

      // in auto trigger mode (no double exposure), put only one image in the image pipe
      if( (LsPar->TRIGGER < 0) && (LsPar->ExposureMode != DOUBLE_EXPOSURE) )
      {
        TakeNextImage  = -1; // next image at next call
      }
      else
      {
        TakeNextImage  =  1; // next image immediately
      }

      calc = capt;
      capt++;
      if (capt == NrOfScreens) capt=0;
    }


    // if it is the last image, don't put another one in the image pipe
    if(LsPar->NBR_LINES == 1)
    {
      if(LsPar->ExposureMode != DOUBLE_EXPOSURE)
      {
        TakeNextImage  = 0;
      }
      else
      {
        if(LsPar->DoubleShutterImg == 1) TakeNextImage  = 0;
      }
    }


    // ready to take next image?
    if(TakeNextImage > 0)
    {
      TakeNextImage = 0;

      // in double exposure mode, take the second image immediately (don't wait for a trigger signal)
      if((LsPar->ExposureMode == DOUBLE_EXPOSURE) && (LsPar->DoubleShutterImg == 1) ) Trigger = 0;

      if( LsPar->ExposureMode == DOUBLE_EXPOSURE) Sh = LsPar->DoubleShutter[capt];
      else                                        Sh = LsPar->SHUTTER;

      shutter(Sh);

      TrackNr[capt] = capture_request(getvar(EXPCNT), getvar(GAIN), (I32 *)Screen[capt], Trigger);

      if(LsPar->DebugPrint == 4) print("Capt. Img X: t=%3dms  CapPage=0x%X  Sh=%4d  Trig=%d  Track[%d]=%3d  DoubleShutterImg=%d  NBR_LINES=%d\n", getvar(MSEC), Screen[capt], Sh, Trigger, capt, TrackNr[capt], LsPar->DoubleShutterImg, LsPar->NBR_LINES);

      // queue is full (this error shoud not happen)
      if(TrackNr[capt] == 0) return SCREEN_RES_ERR_PIPE;
    }


    // important for mode auto trigger (no double exposure)
    if(TakeNextImage < 0) TakeNextImage = 1;


    // image ready for calculation?
    DeltaImg = TrackNr[calc] - getvar(IMGREADY);
    if( (DeltaImg <= 0) || (DeltaImg > NrOfScreens) )
    {
      TakeNextImage = 1;

      ScrSetLogPage(Screen[calc]);

      if(LsPar->DebugPrint == 4) print("Ready Image: t=%3dms  LogPage=0x%X\n", getvar(MSEC), ScrGetLogPage);

      calc = capt;
      capt++;
      if (capt == NrOfScreens) capt=0;

      return SCREEN_RES_IMAGE;
    }

    return SCREEN_RES_BUSY;
  }


  // release screen allocation mode
  if (mode == SCREEN_BUFF_DEINI)
  {
    ScrSetCaptPage(Screen[0]);

    while(cancel_capture_rq());

    // switch back to full image size
    LsPar->ROI_X  =  0;   
    LsPar->ROI_Y  =  0;   
    LsPar->ROI_DX =  ScrGetColumns;   
    LsPar->ROI_DY =  ScrGetRows;   
    SetImageFormat(LsPar);    

    tpict();

    if(ScrAdr) DRAMPgFree(ScrAdr);
  }

  return SCREEN_RES_END;
}

/*************************************************************************************************************************/

I32 AutoShutter (LsParameter *LsPar)
{
  I32 i, Cnt=0, Nr;
  U32 AvgVal, NewSh, AutoShLast;
  U8  *restrict pVal;
  static  I32 AutoShLast2, AutoShLast1 = -1;

  // presetting
  if(AutoShLast1 < 0)  {AutoShLast2 = AutoShLast1 = LsPar->SHUTTER;}

  // calc new shutter time
  Nr = LsPar->LaserPoints;

  pVal = LsPar->pLaserVal;
  for(i=0; i<Nr; i++) if(*pVal++) Cnt++;

  // for(i=0; i<Nr; i++) if(LsPar->pLaserVal[i]) print("%3d: %3d\n", i, (I32)LsPar->pLaserVal[i]);

  if(Cnt > 8)
  {
    // Use detected laser points for brightness calculation.
    AvgVal = FL_sum_U8P(Nr, LsPar->pLaserVal) / Cnt;
  }
  else
  {
    // No laser detected. Use image brightness in order to define, if image is to bright or to dark for laser detection
    image ImgArea;

    ImageAssign(&ImgArea, ScrByteAddr(0, 0), LsPar->ROI_DX, LsPar->ROI_DY_SubSample, ScrGetPitch);

    AvgVal = mean(&ImgArea); if(AvgVal > LsPar->RlcThresh) AvgVal = 255; else AvgVal=0;
  }

  AutoShLast = LsPar->SHUTTER;

  // The 6210 model doesn't change the shutter immediately. It takes one additional image.
  if(getvar(MODEL) == 6210) AutoShLast = AutoShLast2;

  if(AvgVal) NewSh = (LsPar->AutoShutterVal * AutoShLast) / AvgVal;
  else       NewSh =  LsPar->AutoShutterMax;

  if(NewSh < LsPar->AutoShutterMin) NewSh = LsPar->AutoShutterMin;
  if(NewSh > LsPar->AutoShutterMax) NewSh = LsPar->AutoShutterMax;

  if(LsPar->DebugPrint == 3) print("Shutter: Last2=%5d Last1=%5d Last=%5d New=%5d us  /  Val = %3d (%d)  /  Cnt = %4d\n", AutoShLast2, AutoShLast1, LsPar->SHUTTER, NewSh, AvgVal, LsPar->AutoShutterVal, Cnt);

  AutoShLast2 = AutoShLast1;
  AutoShLast1 = LsPar->SHUTTER;

  LsPar->SHUTTER  = NewSh;

  return 0;
}

/************************************************************************************************************************/

I32 AutoTrigger (LsParameter *LsPar, I32 AutoMode)
{
  I32 TimeDiffUs;
  I64 TimeUs;
  static I64 AutoTrigTimeUs;


  // calc absolute time in s
  TimeUs = (I64)1000000 * getlvar(SEC) + 1000 * getvar(MSEC) + TIMER0_DIG / TIMER0_US_DIV;


  if(AutoMode == AUTO_TRIG_INI)
  {
    AutoTrigTimeUs = TimeUs + (10000000 / LsPar->AutoTriggerFPS);

    if(LsPar->DebugPrint == 4) print("AutoTrig   INI: Time=%8dus  NextTrig=%8dus  Err=%3d  (NrLines=%3d)\n", (I32)(TimeUs % 100000000), (I32)(AutoTrigTimeUs % 100000000), LsPar->AutoTriggerError, LsPar->NBR_LINES);
  }


  if(AutoMode == AUTO_TRIG_DIFF)
  {
    TimeDiffUs = (I32)(TimeUs - AutoTrigTimeUs);

    // allow other tasks or interrupts
    if(TimeDiffUs < -1200) time_delay(1);

    // start trigger image
    if(TimeDiffUs >= -50)
    {
      // wait for the exact AutoTrigTimeUs
      while (TimeDiffUs < 0)
      {
        TimeUs = (I64)1000000 * getlvar(SEC) + 1000 * getvar(MSEC) + TIMER0_DIG / TIMER0_US_DIV;
        TimeDiffUs = (I32)(TimeUs - AutoTrigTimeUs);
      }

      AutoTrigTimeUs += (10000000 / LsPar->AutoTriggerFPS);

      if (TimeDiffUs > MAX_JITTER_US) LsPar->AutoTriggerError++;

      if(LsPar->DebugPrint == 4) print("AutoTrig READY: Time=%8dus  NextTrig=%8dus  Err=%3d  (NrLines=%3d / TimeDiff=%3d)\n", (I32)(TimeUs % 100000000), (I32)(AutoTrigTimeUs % 100000000), LsPar->AutoTriggerError, LsPar->NBR_LINES, TimeDiffUs);

      return AUTO_TRIG_READY;
    }
  }

  return 0;
}

/*************************************************************************************************************************/

I32 StoreImage (void)
{
  static I32 ImgCount=0;
  I32    Error;
  char   ImgName[80];
  image  Source;

  ImageAssign(&Source, (U8 *)ScrGetLogPage, ScrGetColumns, ScrGetRows, ScrGetPitch);
  
  { 
    ChangeValue_Telnet(&ImgCount, 0, 999, 1, "Write image number");
    
    // #define JPG_IMG
    
    #ifdef JPG_IMG
      ////////////////////
      // write jpg image
      ////////////////////

      sprint(ImgName, "fd:/LS%03d.jpg",    ImgCount++);
      print("Write image '%s'. Please wait...\n", ImgName);
      Error = fwrite_jpeg(&Source, ImgName, 50, (I32)mremain());
    #else
      ////////////////////
      // write bmp image
      ////////////////////

      sprint(ImgName, "fd:/LS%03d.bmp",    ImgCount++);
      print("Write image '%s'. Please wait...\n", ImgName);
      Error = fwrite_image(ImgName, &Source); 
    #endif

    if (Error) print("Write image Error %d.\n", Error);
    else       print("Write image ready.\n");
  }
  
  return Error;
}

/*************************************************************************************************************************/

I32 ReadImage (void)
{
  static I32 ImgCount=0;
  I32  Error;
  char ImgName[80];
  image Source;

  // take image
  tpict();

  // delete old image
  ImageAssign(&Source, (U8 *)ScrGetLogPage, ScrGetColumns, ScrGetRows, ScrGetPitch);
  set(&Source, 0);

  // if (DmPar->DmError != 0)
  { 
    ChangeValue(&ImgCount, 0, 999, 1, "Read image number");

    sprint(ImgName, "fd:/LS%03d.bmp", ImgCount);
    print("Read image '%s'. Please wait...\n", ImgName);

    ////////////////////
    // read bmp image
    ////////////////////

    Error = fread_image(ImgName, &Source); 
     
    if (Error == 0) 
    {
      print("Read image ready.\n");
    }
    else
    {
      if (Error == ERR_OPEN) print("File '%s' not found.\n", ImgName);
      else                   print("Read BMP Typ Error %d.\n", Error);
    }
  }

  return Error;
}

/***************************************************************************************************************************/

void SensorReset (LsParameter *LsPar)
{
  // shutter(LsPar->SHUTTER);
  setvar(GAIN, LsPar->GAINVAL);

  // set laser
  if((getvar(MODEL) / 100) == 62) SwitchLaser(LsPar->LASER_MODE);

  // reset double exposer image
  LsPar->DoubleShutterImg = 0;

  // set trigger mode
  if(LsPar->TRIGGER == TRIGGER_RISING_EDGE ) TRIGINP_POS();
  if(LsPar->TRIGGER == TRIGGER_FALLING_EDGE) TRIGINP_NEG();

  // RESET_SENSOR (every cancel_capture_rq() will reset the image format!)
  ScreenBuffer2(LsPar, SCREEN_BUFF_RESET);

  // set ROI range in mm
  ImgToWorld_ROI(LsPar);
}

/************************************************************************************************************************/

#define C6210_LCON              0x44000000          /* serial shift register for LCON     */

I32 SwitchLaser (I32 LaserOn)
{
  I32 LaserPower_mW = LASER_POWER;

  if(LaserOn)
  {
    // printf("switch on laser\n");

    if(LaserPower_mW == 100)
    {
      // set current
      *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg */
      *((volatile char *)C6210_LCON) = 0xFF;  /* data register    */
      *((volatile char *)C6210_LCON) = 0x00;  /* address register */

      time_delay(2);

      // set channel
      *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg */
      *((volatile char *)C6210_LCON) = 0x1F;  /* control register   */
      *((volatile char *)C6210_LCON) = 0x01;  /* address register   */
    }

    if(LaserPower_mW == 30)
    {
      *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg    */
      *((volatile char *)C6210_LCON) = 0x80;  /* current register    */
      *((volatile char *)C6210_LCON) = 0x00;  /* address register    */

      time_delay(200);

      *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg    */
      *((volatile char *)C6210_LCON) = 0x1F;  /* control register        */
      *((volatile char *)C6210_LCON) = 0x01;  /* address register        */
    }
  }
  else
  {
    // printf("switch off laser\n");

    // set current
    *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg */
    *((volatile char *)C6210_LCON) = 0x00;  /* data register    */
    *((volatile char *)C6210_LCON) = 0x00;  /* address register */

    time_delay(2);

    // set channel
    *((volatile char *)C6210_TTL) = 0;      /* write access to different register clears aflg */
    *((volatile char *)C6210_LCON) = 0x00;  /* control register   */
    *((volatile char *)C6210_LCON) = 0x01;  /* address register   */
  }

  return 0;
}

/***************************************************************************************************************************/

I32 SetImageFormat (LsParameter *LsPar)
{
  I32 SubsamplingFactor=1, mode=0; // 0=bw 1=color sensor
  I40 NewShutter;

  LsPar->ROI_Y_SubSample  = LsPar->ROI_Y; 
  LsPar->ROI_DY_SubSample = LsPar->ROI_DY; 

  if( (getvar(MODEL) / 100) == 62)
  {
    // print("Set Image Format (SubSample=%d) X=%3d Y=%3d DX=%d DY=%3d SH=%4d -> ", LsPar->SubSample, LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->SHUTTER);

    // set image format (every cancel_capture_rq() will reset the format!)
    if(getvar(MODEL) == 6211)
    {
      if(LsPar->SubSample) SubsamplingFactor = (1 << LsPar->SubSample);
    
      set_image_format_ev76c560_subs(SubsamplingFactor, LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->SHUTTER, mode, &LsPar->ROI_X, &LsPar->ROI_Y, &LsPar->ROI_DX, &LsPar->ROI_DY, &NewShutter); 
      
      LsPar->ROI_Y_SubSample  = LsPar->ROI_Y  / SubsamplingFactor; 
      LsPar->ROI_DY_SubSample = LsPar->ROI_DY / SubsamplingFactor; 
    }
    else
    {
      vcrt_set_image_format(LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->SHUTTER, mode, &LsPar->ROI_X, &LsPar->ROI_Y, &LsPar->ROI_DX, &LsPar->ROI_DY, &NewShutter);

      LsPar->SubSample        = 0;
      LsPar->ROI_Y_SubSample  = LsPar->ROI_Y; 
      LsPar->ROI_DY_SubSample = LsPar->ROI_DY; 
    }

    // print("X=%3d Y=%3d DX=%d DY=%3d SH=%4d\n", LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, (I32)NewShutter);
  }

  return 0;
}

/************************************************************************************************************************/

/************************************************************************************************************************/
//  TCP / IP - FUNCTIONS
/************************************************************************************************************************/

/************************************************************************************************************************/

I32 SetCommandPar(LsParameter *LsPar, I32 *Variable, I32 *MinVal, I32 *MaxVal, I32 *NewVal, I32 Action)
{
  I32 SetVal, Error = ERR_PARM;

  // get range
  GetParameterRange(LsPar, Variable, &SetVal, MinVal, MaxVal);

  if( (Action == 0) && (*NewVal >= *MinVal) && (*NewVal <= *MaxVal) )
  {
    Error = ERR_NONE;

    // set parameter
    *Variable = *NewVal;
  }

  *NewVal = *Variable;

  return Error;
}

/************************************************************************************************************************/

I32 TCP_RunCommand (LsParameter *LsPar, U8 *Command, I32 length, U32 *Alias)
{
  I32   i, Error, CommandError = ERR_TCP_IP;
  I32   MinVal, MaxVal, NewVal;
  I32   CommandID, Action;
  U32   HostCounterNew;
  U8    *pCom, *pEnd;



  /////////////////////////////////////////////////////////
  // ini values
  ///////////////////////////////////////////////////////// 

  Error       =  0;
  CommandID   =  CMD_ERROR;
  Action      = -1;
  pCom        = Command;
  pEnd        = Command + length + 1;


  // print host string
  if (LsPar->PrintHostCommand & 0x01)
  {
    // for(i=0; i<length; i++) print("CommandRead[%4d / %4d] = %c\n", i, length, Command[i]);
  }


  NEXT_COMMAND:

  /////////////////////////////////////////////////////////
  // Read Command: <CmdID> <HostCounter> <Action> <Val>
  /////////////////////////////////////////////////////////

  // ini settings
  MinVal = 0;
  MaxVal = 0;
  NewVal = 0;
  
  if(LsPar->CommandModeASCII)
  {
    I8 Val[9];
    
    // set end of string
    Val[8] = 0;
    
    // read CommandID
    FL_CPY_U8P_U8P(8, &pCom[ 0], (U8 *)Val);
    CommandID = atoi(Val);
    // print("String %s -> Int %d\n", Val, atoi(Val));

    // read HostCounter
    FL_CPY_U8P_U8P(8, &pCom[ 8], (U8 *)Val);
    HostCounterNew = atoi(Val);
    // print("String %s -> Int %d\n", Val, atoi(Val));

    // read Action
    FL_CPY_U8P_U8P(8, &pCom[16], (U8 *)Val);
    Action = atoi(Val);
    // print("String %s -> Int %d\n", Val, atoi(Val));

    // read value
    FL_CPY_U8P_U8P(8, &pCom[24], (U8 *)Val);
    NewVal = atoi(Val);
    // print("String %s -> Int %d\n", Val, atoi(Val));

    // pointer to the end of I32 read values
    pCom += (4*8);
  }
  else
  {
    // use I32 pointer to read I32 values
    I32 *pI32 = (I32 *)pCom;
  
    // read CommandID
    CommandID = *pI32++;

    // read HostCounter
    HostCounterNew = *pI32++;

    // read Action
    Action = *pI32++;

    // read value
    NewVal = *pI32++;

    // pointer to the end of I32 read values
    pCom = (U8 *)pI32;
  }

  // print host string
  if (LsPar->PrintHostCommand & 0x01)
  {
    print("Host->Cam: ");
    print("CommandID=%2d  HostCounter=%3d [Old=%3d]  Action=%d  NewVal=%4d  Length=%4d\n", CommandID, HostCounterNew, HostCounter, Action, NewVal, length);
  }

  // send remaining data before acknowledge new command
  if(CommandID == CMD_STOP) TCP_SendCommand(LsPar, Alias, LsPar->SendBuffer, LsPar->MaxSendBuff, INI_SEND_BUFFER);
  else                      TCP_SendCommand(LsPar, Alias, LsPar->SendBuffer, LsPar->MaxSendBuff, EMPTY_SEND_BUFFER);


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // HOST -> VISION
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  switch(CommandID)
  {
    case  CMD_STOP:
          HostCounter = 0; 
          CamCounter  = 0; 
          LineCounter = 0;
          CommandError = SetCommandPar(LsPar, &LsPar->NBR_LINES, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_LASER_MODE:
          CommandError = SetCommandPar(LsPar, &LsPar->LASER_MODE, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ROI_X:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_X, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ROI_Y:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_Y, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ROI_DX:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DX, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ROI_DY:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DY, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_GAIN:
          CommandError = SetCommandPar(LsPar, &LsPar->GAINVAL, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_SHUTTER:
          CommandError = SetCommandPar(LsPar, &LsPar->SHUTTER, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_TRIGGER:
          CommandError = SetCommandPar(LsPar, &LsPar->TRIGGER, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_MODE:
          CommandError = SetCommandPar(LsPar, &LsPar->MODE, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_NBR_LINES:
          CommandError = SetCommandPar(LsPar, &LsPar->NBR_LINES, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_METHOD:
          CommandError = SetCommandPar(LsPar, &LsPar->METHOD, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ALGORITHM:
          CommandError = SetCommandPar(LsPar, &LsPar->ALGORITHM, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_THRESH:
          CommandError = SetCommandPar(LsPar, &LsPar->RlcThresh, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_MINWIDTH:
          CommandError = SetCommandPar(LsPar, &LsPar->MinWidth, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_MAXWIDTH:
          CommandError = SetCommandPar(LsPar, &LsPar->MaxWidth, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_EXPMODE:
          CommandError = SetCommandPar(LsPar, &LsPar->ExposureMode, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_LASERVAL:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterVal, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_MINSH:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterMin, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_MAXSH:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterMax, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_DOUBLE_SH1:
          CommandError = SetCommandPar(LsPar, &LsPar->DoubleShutter[0], &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_DOUBLE_SH2:
          CommandError = SetCommandPar(LsPar, &LsPar->DoubleShutter[1], &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_FILTER:
          CommandError = SetCommandPar(LsPar, &LsPar->ReflexionFilter, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_ROI_X_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_X_MM, &MinVal, &MaxVal, &NewVal, Action);
          if(Action == 0) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_Y_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_Y_MM, &MinVal, &MaxVal, &NewVal, Action);
          if(Action == 0) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_DX_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DX_MM, &MinVal, &MaxVal, &NewVal, Action);
          if(Action == 0) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_DY_MM:
          if(Action == 0) WorldToImg_ROI(LsPar);
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DY_MM, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_LIB_VERSION:
          CommandError = SetCommandPar(LsPar, &LsPar->LibVersion, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_PLC_IN:
          NewVal = inPLC() & 0x0F;
          CommandError = SetCommandPar(LsPar, &LsPar->InPLC, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_PLC_OUT:
          CommandError = SetCommandPar(LsPar, &LsPar->OutPLC, &MinVal, &MaxVal, &NewVal, Action);
          if(CommandError == ERR_NONE) outPLC(NewVal);
          break;

    case  CMD_ETHERNET_PACKAGE:
          CommandError = SetCommandPar(LsPar, &LsPar->EthernetPackNr, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_AUTO_TRIG_PERIODE:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoTriggerFPS, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_AUTO_TRIG_ERR:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoTriggerError, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_SAVE_SETTINGS:
          NewVal = StoreSensorParameter(LsPar);
          break;

    case  CMD_SEND_NOWAIT:
          CommandError = SetCommandPar(LsPar, &LsPar->EthernetSendNoWait, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_SUBSAMPLE:
          CommandError = SetCommandPar(LsPar, &LsPar->SubSample, &MinVal, &MaxVal, &NewVal, Action);
          break;

    case  CMD_CAMERA_SNR:
          CommandError = SetCommandPar(LsPar, &LsPar->CameraSNR, &MinVal, &MaxVal, &NewVal, Action);
          break;
          
    #ifdef PRODUCT_VERSION
    case  CMD_READ_PRODUCT_FILE:
          NewVal = 0;
          CommandError = ERR_NONE;
          IniProductParameter(LsPar, READ_DATA);
          break;
          
    case  CMD_SET_PRODUCT_DATA:
          CommandError = ERR_NONE;
          break;
          
    case  CMD_SEND_PRODUCT_DATA:
          NewVal = 0;
          CommandError = ERR_NONE;
          break;
    #endif
          
    case  CMD_READ_SENSOR_FILE:
          NewVal = 0;
          CommandError = ERR_NONE;
          InitSensorParameter(LsPar, READ_DATA);
          break;
          
    case  CMD_SET_SENSOR_DATA:
          CommandError = ERR_NONE;
          break;
          
    case  CMD_SEND_SENSOR_DATA:
          NewVal = 0;
          CommandError = ERR_NONE;
          break;
          
    default:
          NewVal = 0;
          CommandError = ERR_COMMAND;
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // command response
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  {
    I32 Response[64 + MAX_SETTING_SIZE/4], *pResponse, SendLen;

    // check HostCounter
    if(HostCounter >= 99999999) HostCounter = 0;
    if(HostCounterNew != (U32)(HostCounter + 1)) CommandError = ERR_COUNTER;
    HostCounter = HostCounterNew;

    pResponse = Response;

    if(Action)
    {
      ////////////////////
      // read mode
      ////////////////////

      if(LsPar->CommandModeASCII)
      {
        I8 *pI8 = (I8 *)pResponse;
        
        // set "Response Global Header" + "Response Meta Header"
        sprint(pI8, "%8d%8d%8d%8d%8d%8d%8d%8d", 100, CamCounter, 4*8, CommandID, MinVal, MaxVal, NewVal);
        pResponse += 7 * 8 / 4; // I32 values
      }
      else
      {
        // set "Response Global Header"
        *pResponse++ = 100;
        *pResponse++ = CamCounter;
        *pResponse++ = 16;

        // set "Response Meta Header"
        *pResponse++ = CommandID;
        *pResponse++ = MinVal;
        *pResponse++ = MaxVal;
        *pResponse++ = NewVal;
      }
    }
    else
    {
      ////////////////////
      // set mode
      ////////////////////

      if(LsPar->CommandModeASCII)
      {
        I8 *pI8 = (I8 *)pResponse;
        
        // set "Response Global Header" + "Response Meta Header"
        sprint(pI8, "%8d%8d%8d%8d%8d%8d", 101, CamCounter, 2*8, CommandID, CommandError);
        pResponse += 5 * 8 / 4; // I32 values
      }
      else
      {
        // set mode

        // set "Response Global Header"
        *pResponse++ = 101;
        *pResponse++ = CamCounter;
        *pResponse++ = 8;

        // set "Response Meta Header"
        *pResponse++ = CommandID;
        *pResponse++ = CommandError;
      }
    }
    
    CamCounter++;
    if(CamCounter > 99999999) CamCounter = 0;

    // Send response
    SendLen = 4 * (I32)(pResponse - Response);

    // special command to send sensor parameter from Cam -> Host
    if(CommandID == CMD_SEND_SENSOR_DATA)
    {
      I32 Size = LsPar->SensorSettingSize;
      
      Response[2] += Size; // set new length
      SendLen     += Size; // set new length
      FL_CPY_U8P_U8P(Size, (U8 *)LsPar->pSensorSetting, (U8 *)pResponse);
    }

    // special command to send sensor parameter from Host -> Cam
    if(CommandID == CMD_SET_SENSOR_DATA)
    {
      I32 Size = NewVal;
      
      if(pCom + Size < pEnd)
      {
        FL_CPY_U8P_U8P(Size, pCom, (U8 *)LsPar->pSensorSetting);

        pCom                    += Size;
        LsPar->SensorSettingSize = Size;
        
        InitSensorParameter(LsPar, SET_DATA);
      }
      
      NewVal = 0;
    }

// print("CMD_ID=%d Action=%d\n", CommandID, Action); 

    #ifdef PRODUCT_VERSION
    
    // special command to send product parameter from Cam -> Host
    if(CommandID == CMD_SEND_PRODUCT_DATA)
    {
      I32 Size = ProdPar->ProdSettingSize;
      
      Response[2] += Size; // set new length
      SendLen     += Size; // set new length
      FL_CPY_U8P_U8P(Size, (U8 *)ProdPar->pProductSetting, (U8 *)pResponse);
    }

    // special command to send product parameter from Host -> Cam
    if(CommandID == CMD_SET_PRODUCT_DATA)
    {
      I32 Size = NewVal;
      
      if(pCom + Size < pEnd)
      {
        FL_CPY_U8P_U8P(Size, pCom, (U8 *)ProdPar->pProductSetting);

        pCom                    += Size;
        ProdPar->ProdSettingSize = Size;
        
        SetProductParameter(LsPar);
      }
      
      NewVal = 0;
    }
    
    #endif

    while(SendLen > 0)
    {
      I32 Len = TCP_Send(*Alias, (char *)Response, SendLen);

      if(Len > 0) SendLen -= Len;
      else        SendLen  =  -1;

      // print("SendLen=%d Len=%d\n", SendLen, Len);
    }


    // print response string
    if (LsPar->PrintHostCommand & 0x02)
    {
      I8 *pI8 = (I8 *)Response;
      
      print("Cam->Host: ");

      if(LsPar->CommandModeASCII) for (i=0; i<4 * (I32)(pResponse - Response); i++) print("%c", *pI8++);
      else                        for (i=0; i<    (I32)(pResponse - Response); i++) print("%d ", Response[i]);

      print("\n\n");
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // More Commands?
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // if(pCom + sizeof(I32) * LsPar->MinReceiveBuff < pEnd) print("NEXT_COMMAND\n");
  if(pCom + sizeof(I32) * LsPar->MinReceiveBuff < pEnd) goto NEXT_COMMAND;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // RESET_SENSOR (every cancel_capture_rq() will reset the image format!)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if((Action == 0) && (CommandID != CMD_NBR_LINES) && (CommandID != CMD_MODE) ) SensorReset(LsPar);

  return Error;
}

/************************************************************************************************************************/

U8 *CreateSendHeader (U8 *pWrite, I32 FormatASCII, I32 Mode, I32 CamCounter, I32 Length)
{
  I32 Size, *pI32;
  
  // print("%8d%8d%8d\n", Mode, CamCounter, Length);
  
  if(FormatASCII)
  {
    sprint((I8 *)pWrite, "%8d%8d%8d", Mode, CamCounter, Length);
    Size = 3 * 8;
  }
  else
  { 
    pI32 = (I32 *)pWrite;
    
    *pI32++ = Mode;
    *pI32++ = CamCounter;
    *pI32++ = Length;

    Size = 3 * sizeof(I32);
  }
  
  return (pWrite + Size);
}

/************************************************************************************************************************/

I32 TCP_SendCommand (LsParameter *LsPar, U32 *Alias, U8 *Command, I32 maxlen, I32 BufferMode)
{
  I16    x, y, dx, dy;
  I32    i, count, SendSize, DataSize, BuffLen, Error=0;
  I32    MaxHeaderSize, *pI32;
  static U8  *restrict pSend  = NULL;
  static U8  *restrict pWrite = NULL;
  static I32 PackCnt=0, OptNoWait=1;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // ini values
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if(BufferMode ==   INI_SEND_BUFFER) {pSend = NULL; return 0;}
  if(BufferMode == EMPTY_SEND_BUFFER) {if( (I32)(pWrite - pSend) == 0) return 0;}


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // ini values
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if(pSend == NULL) {pSend = pWrite = Command;}

  BuffLen = maxlen - (I32)(pWrite - Command);

  // print("NBR_LINES=%3d  CamCounter=%3d  HostCounter=%3d\n", LsPar->NBR_LINES, CamCounter, HostCounter);

  // no TCP/IP transfer
  if(LsPar->MODE < 0) return 0;

  // check ethernet cable connection
  if(getvar(ETHLINK) == 0) return 0;



  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // check parameter
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (LsPar->MODE < 1) || (LsPar->MODE > 9) ) return ERR_PARM;

  // print("LsPar->MODE = %d\n", LsPar->MODE);


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // calc size
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  DataSize      =  0;
  MaxHeaderSize = 40; // Byte

  if(LsPar->MODE == 1) DataSize = MaxHeaderSize +     LsPar->ROI_DX * LsPar->ROI_DY_SubSample;
  if(LsPar->MODE == 2) DataSize = MaxHeaderSize +     LsPar->LaserPoints * sizeof(float);
  if(LsPar->MODE == 3) DataSize = MaxHeaderSize +     LsPar->LaserPoints * (sizeof(float) + 1);
  if(LsPar->MODE == 4) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float);
  if(LsPar->MODE == 5) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float);
  if(LsPar->MODE == 6) DataSize = MaxHeaderSize +     LsPar->ROI_DX * LsPar->ROI_DY_SubSample;
  if(LsPar->MODE == 7) DataSize = MaxHeaderSize +     LsPar->LaserPoints * (sizeof(float) + 1) + 256;
  if(LsPar->MODE == 8) DataSize = MaxHeaderSize +     256;
  if(LsPar->MODE == 9) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float) + 2000;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // optimize TPP/IP transfer (OPT_SEND_NOWAIT)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  {
    /* send data immediately (VCRT_MSG_NONBLOCK) */
    I32 opt_value = 1;

    if(LsPar->EthernetSendNoWait  == 0) opt_value = 0;
    if(LsPar->MODE                == 1) opt_value = 0;
    if(LsPar->MODE                == 6) opt_value = 0;
    if(BuffLen     <          DataSize) opt_value = 0;
    if(BufferMode == EMPTY_SEND_BUFFER) opt_value = 0;

    if(opt_value != OptNoWait)
    {
      OptNoWait = opt_value;

      Error = setsockopt(*Alias, SOL_TCP, OPT_SEND_NOWAIT, &opt_value, sizeof(opt_value));

      // print("New opt_value = %d  Error=%d\n", opt_value, Error);
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // if not enough memory for new data, send complete buffer
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (BuffLen < DataSize) || (BufferMode == EMPTY_SEND_BUFFER) )
  {
    // calc response len
    count = (I32)(pWrite - pSend);

    // send all data (opt_value must be 0 !)
    SendSize = TCP_Send(*Alias, (char *)pSend, count);

    if (LsPar->DebugPrint == 5) print("Buff SendSize=%6d Remain=%10d (Empty Buffer)\n", SendSize, count - SendSize);

    // print("Error TCP Send: Buffer Full\n");

    pSend = pWrite = Command;

    if(SendSize < 0) {if (LsPar->DebugPrint == 5) print("Error=%d TCP Send: Size=%d\n", SendSize, count); return -1;}
    if(BufferMode == EMPTY_SEND_BUFFER) {return 0;}
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // VISION -> HOST (write SendDate in Buffer)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  0. State MODE_STANDBY
  1. State MODE_TAKE_PIC
  2. State MODE_POINTLIST
  3. State MODE_POINTLIST_AND_LASER_IMG
  4. State MODE_POINTLIST_MM_XZ_XZ (x0 z0 / x1 z1 / x2 z2 / ...)
  5. State MODE_POINTLIST_MM_XX_ZZ (x0 x1 x2 ... / z0 z1 z2 ...)
  6. State MODE_BINARY_PIC
  7. State MODE_POINTLIST_AND_LASER_IMG + Product Angle Results
  8. State MODE_PRODUCT_RESULTS
  9. State MODE_POINTLIST_MM + Product Angle Results
  */

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 1. State MODE_TAKE_PIC
  // 6. State MODE_BINARY_PIC
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (LsPar->MODE == 1) || (LsPar->MODE == 6) )
  {
    I32 NextLine;
    U8  *restrict pPix;

    // set "Response Global Header"
    pWrite = CreateSendHeader(pWrite, LsPar->CommandModeASCII, LsPar->MODE, CamCounter, 8);

    // set "Response Meta Header"
    pI32    = (I32 *)pWrite;
    *pI32++ =  LsPar->ROI_DX;
    *pI32++ =  LsPar->ROI_DY_SubSample;
    pWrite  = (U8  *)pI32;

    // copy image data
    pPix     = (U8 *)ScrGetLogPage;
    dx       =       LsPar->ROI_DX;
    dy       =       LsPar->ROI_DY_SubSample;
    NextLine =       ScrGetPitch - dx;

    for(y=0; y<dy; y++)
    {
      if(LsPar->MODE == 1)
      {
        for(x=0; x<dx; x++)
        {
          *pWrite++ = *pPix++;
        }
      }
      else
      {
        for(x=0; x<dx; x++)
        {
          if(*pPix++ >= LsPar->RlcThresh) *pWrite++ = 255;
          else                            *pWrite++ =   0;
        }
      }

      pPix += NextLine;
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 2. State MODE_POINTLIST
  // 3. State MODE_POINTLIST_AND_LASER_IMG
  // 7. State MODE_POINTLIST_AND_LASER_IMG + Product Angle Results
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (LsPar->MODE == 2) || (LsPar->MODE == 3) || (LsPar->MODE == 7) )
  {
    I32 Nr=0;

    // set "Response Global Header"
    pWrite = CreateSendHeader(pWrite, LsPar->CommandModeASCII, LsPar->MODE, CamCounter, 16);

    // set "Response Meta Header"
    pI32    = (I32 *)pWrite;
    *pI32++ =  LsPar->ALGORITHM;
    *pI32++ =  LsPar->ROI_DX / LsPar->StepDx;
    *pI32++ =  LsPar->ROI_X;
    *pI32++ =  LsPar->StepDx;
    pWrite  = (U8  *)pI32;
    
    // set "Response DATA"
    if(LsPar->ALGORITHM)
    {
      float *pFloat = (float *)(pWrite), Accuracy_Inv = 1.0 / (float)LASER_POS_MUL;

      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to NO_LASER_DETEC
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) *pFloat++ = (float)LsPar->pLaserPosZ[Nr++] * Accuracy_Inv;
        else                                                                  *pFloat++ = (float)NO_LASER_DETEC          * Accuracy_Inv;
      }

      pWrite = (U8 *)pFloat;
    }
    else
    {
      I16 *pI16 = (I16 *)(pWrite);

      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to NO_LASER_DETEC
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) *pI16++ = (I16)(LsPar->pLaserPosZ[Nr++] >> 2);
        else                                                                  *pI16++ = (I16)(NO_LASER_DETEC          >> 2);
      }

      pWrite = (U8 *)pI16;
    }

    if( (LsPar->MODE == 3) || (LsPar->MODE == 7) )
    {
       Nr = 0;
       
      // ADD Laser Brightness to "Response DATA"
      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to color 0
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) *pWrite++ = LsPar->pLaserVal[Nr++];
        else                                                                  *pWrite++ = 0;
      }
    }
    

    #ifdef PRODUCT_VERSION
    if(LsPar->MODE == 7)
    {
      // Product Angle Results
      {
        I32 Nr       = (I32)((U8 *)(&ProdPar->AngleError) - (U8 *)(&ProdPar->TargetAngle)) + 4;
        U8  *pStruct = (U8 *)&ProdPar->TargetAngle;
        U8  *pTemp   = (U8 *)&Nr;

        for(i=0; i< 4; i++) *pWrite++ = *pTemp++;
        for(i=0; i<Nr; i++) *pWrite++ = *pStruct++;
      }
    }
    #endif
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 4. State MODE_POINTLIST_MM (x0 z0 / x1 z1 / x2 z2 / ...)
  // 5. State MODE_POINTLIST_MM (x0 x1 x2 ... / z0 z1 z2 ...)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (LsPar->MODE == 4) || (LsPar->MODE == 5) )
  {
    I32   CountPoints = LsPar->LaserPoints;
    float *pWorld;

    // set "Response Global Header"
    pWrite = CreateSendHeader(pWrite, LsPar->CommandModeASCII, LsPar->MODE, CamCounter, 8);

    // set "Response Meta Header"
    pI32    = (I32 *)pWrite;
    *pI32++ =  CountPoints;
    *pI32++ =  LineCounter++;
    pWrite  = (U8  *)pI32;
    
    pWorld = (float *)pWrite;

    // write world coordinates into transfer buffer
    if(LsPar->MODE == 4)
    {
      for(i=0; i<CountPoints; i++) 
      {
        *pWorld++ = LsPar->pLaserPosX_MM[i];
        *pWorld++ = LsPar->pLaserPosZ_MM[i];
      }
    }

    if(LsPar->MODE == 5)
    {
      for(i=0; i<CountPoints; i++) *pWorld++ = LsPar->pLaserPosX_MM[i];
      for(i=0; i<CountPoints; i++) *pWorld++ = LsPar->pLaserPosZ_MM[i];
    }

    pWrite = (U8 *)pWorld;
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 8. State MODE_PRODUCT_RESULTS (ASCII)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  #ifdef PRODUCT_VERSION
  if( LsPar->MODE == 8 )
  {
    I32  Nr;

    // set "Response Global Header"
    pWrite = CreateSendHeader(pWrite, LsPar->CommandModeASCII, LsPar->MODE, CamCounter, 8);

    // set "Response Meta Header"
    pI32    = (I32 *)pWrite;
    *pI32++ =  LineCounter++;
    Nr      =  strlen(ProdPar->ProductResultString); // set "Response DATA"
    *pI32++ =  Nr;
    pWrite  = (U8  *)pI32;

    // response data
    for(i=0; i<Nr; i++) *pWrite++ = ProdPar->ProductResultString[i];
  }
  #endif


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 9. State MODE_POINTLIST_MM + Product Results (x0 z0 / x1 z1 / x2 z2 / ...) + Product Results (ASCII)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  #ifdef PRODUCT_VERSION
  if( LsPar->MODE == 9 )
  {
    I32   CountPoints = LsPar->LaserPoints;
    I32   ProdResLen  =  strlen(ProdPar->ProductResultString); 
    float *pWorld;

    // set "Response Global Header"
    pWrite = CreateSendHeader(pWrite, LsPar->CommandModeASCII, LsPar->MODE, CamCounter, 12);

    // set "Response Meta Header"
    pI32    = (I32 *)pWrite;
    *pI32++ =  CountPoints;
    *pI32++ =  LineCounter++;
    *pI32++ =  ProdResLen;
    pWrite  = (U8  *)pI32;
    
    pWorld = (float *)pWrite;

    // write world coordinates into transfer buffer
    if(LsPar->MODE == 4)
    {
      for(i=0; i<CountPoints; i++) 
      {
        *pWorld++ = LsPar->pLaserPosX_MM[i];
        *pWorld++ = LsPar->pLaserPosZ_MM[i];
      }
    }

    // transfer response data
    pWrite = (U8 *)pWorld;
    for(i=0; i<ProdResLen; i++) *pWrite++ = ProdPar->ProductResultString[i];
  }
  #endif


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // send MODE (VISIOM -> HOST)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // increase camera counter
  CamCounter++;
  if(CamCounter > 99999999) CamCounter = 0;
  
  PackCnt++;

  // transfer results (or build packages, if MODE = 4 or 5)
  // if ( ((LsPar->MODE != 4) && (LsPar->MODE != 5)) || (PackCnt >= LsPar->EthernetPackNr) || (LsPar->NBR_LINES == 0) )
  if ( (LsPar->NBR_LINES == 0) || ((PackCnt >= LsPar->EthernetPackNr) && (LsPar->EthernetPackNr)) || (LsPar->MODE == 1) || (LsPar->MODE == 6) )
  {
    I32 OptiSendSize;

    // reset package counter
    PackCnt = 0;

    // calc response len
    count = (I32)(pWrite - pSend);

    while (count)
    {
      OptiSendSize = (count / 1460) * 1460;
      if(OptiSendSize > 44 * 1460) OptiSendSize = 44 * 1460;
      if(OptiSendSize == 0) OptiSendSize = count;

      // send full size in BlockingMode
      if(LsPar->MODE == 1 || LsPar->MODE == 6) OptiSendSize = count;

      SendSize = TCP_Send(*Alias, (char *)pSend, OptiSendSize);

      if(SendSize <  0) {if (LsPar->DebugPrint == 5) print("Error=%d TCP Send: Size=%d\n", SendSize, count); pSend = pWrite = Command; return -1;}
      if(SendSize == 0) {time_delay(1);}

      count -= SendSize;
      pSend += SendSize;

      if (LsPar->DebugPrint == 5) print("Main SendSize=%6d Remain=%10d (Empty Buffer)\n", SendSize, count);

      // stop send data, if more scans are requested
      if(LsPar->EthernetSendNoWait)
      {
        if( (LsPar->NBR_LINES != 0) && (LsPar->EthernetPackNr != 0) && (LsPar->MODE != 1) && (LsPar->MODE != 6)) break;
        // if( (LsPar->NBR_LINES != 0) && (LsPar->EthernetPackNr != 0) ) break;
      }

      // if(LoopNr-- < 0) {pSend = pWrite = Command; count = 0;}
    }

    if(count == 0) {pSend = pWrite = Command;}

    // print response string
    if (LsPar->PrintHostCommand & 0x04)
    {
      print("Cam->Host: ");

      if(LsPar->CommandModeASCII) for (i=0; i<48; i++) print("%c", Command[i]);
      else                        for (i=0; i<16; i++) print("%02x ", ((I32)Command[i]) & 0xFF);
      
      print("\n\n");
    }
  }

  return Error;
}

/************************************************************************************************************************/

I32 TCP_ReadCommand (U32 sock, U32 *Alias, char *Byte, I32 minlen, I32 maxlen, I32 *length)
{
  I32  count=0, result, ms, DiffMs, ReadyString=0;

  ms = getvar(MSEC);

  do
  {
    result = TCP_DataAvailable(sock, Alias, (char *)&Byte[count], maxlen);

    // if(result >= 0) print("result=%d\n", result);

    if(result >= 0)
    {
      count  += result;
      maxlen -= result;

      if(count >= minlen)
      {
        // for(i=1; i<=result; i++) if(Byte[count - i] == '#') {ReadyString = 1; break;}
        ReadyString = 1;
      }

      // ms     = getvar(MSEC);
    }
    else
    {
      if(count == 0) break;
    }

    DiffMs = getvar(MSEC) - ms;
    if (DiffMs < 0) DiffMs += 1000;
  }
  while ( (DiffMs < TCP_TIMEOUT) && (ReadyString == 0) );

  *length = count;

  // if(count > 0) print("count=%d\n", count);

  if (ReadyString) return( 0);
  else             return(-1);
}

/************************************************************************************************************************/

I32 TCP_DataAvailable (U32 sock, U32 *Alias, char *Byte, I32 length)
{
  I32             ConnectionTimeOut, result, ReceiveL=0;
  static unsigned newsock=0, ContactingRemoteSocket, AliasSocket;
  static I32      ConnectionExists=0;


  ConnectionTimeOut = -1; // ms

  /* Are there connections at any port? */
  // ContactingRemoteSocket = sock;
  result = TCP_Connection(ConnectionTimeOut, &ContactingRemoteSocket);

  /* check for connection errors */
  if (result)
  {
    // if (result == -1) print ("VCRT Connection Socket Error = %d\n", ContactingRemoteSocket);
    // if (result == -2) print ("Connection Time Out Error\n");
    return(result);
  }

  /* Is there one who wants to connect to port SERVER_PORT */
  if (sock == ContactingRemoteSocket)
  {
    print("New Connection on Server Socket\n");

    // reset Counter
    HostCounter = 0;
    CamCounter  = 0;


    // Close existing connection
    if (ConnectionExists)
    {
      ConnectionExists=0;
      TCP_Close(newsock);
      print("Close Old Socket\n");
    }

    result = TCP_Accept(ContactingRemoteSocket, &AliasSocket);

    if (result)
    {
      print("Accept failed, error 0x%x\n",VCRT_geterror(AliasSocket));
      ConnectionExists=0;
      return(result);
    }
    else
    {
      print("accepted\n");
      print ("Alias Socket=0x%x\n", AliasSocket);
      ConnectionExists=1;
      newsock = AliasSocket;
      *Alias  = AliasSocket;

      ReceiveL = TCP_Receive(newsock, Byte, length);

      if (VCRT_geterror(AliasSocket) != VCRT_OK)
      {
        if(ConnectionExists) TCP_Close(AliasSocket);
        return(-1);
      }
    }
  }
  else
  {
    if (VCRT_geterror(AliasSocket) == VCRT_OK)
    {
      ReceiveL = TCP_Receive(newsock, Byte, length);

      if (VCRT_geterror(AliasSocket) != VCRT_OK)
      {
        if(ConnectionExists) TCP_Close(AliasSocket);
        return(-1);
      }
    }
    else
    {
      if(ConnectionExists) TCP_Close(AliasSocket);
      return(-1);
    }
  }

  return(ReceiveL);
}

/************************************************************************************************************************/

#define TCP_IP_SOCKET

I32 TCP_Open (I32 Port, U32 *NewSock)
{
  unsigned  error, sock;
  I32  Error;
//  unsigned  opt_value=1, opt_length=sizeof(unsigned);
  sockaddr_in  laddr;


  /* server services port */
  laddr.sin_family      = AF_INET;
  laddr.sin_port        = Port;        /* server listen at port */
  laddr.sin_addr.s_addr = INADDR_ANY;  /* every IP addr can connect */


  /* use TCP/IP protocoll and search for a free port */
  #ifdef TCP_IP_SOCKET
  sock = socket_stream();
  #else
  sock = socket_dgram();
  #endif

  if (sock == VCRT_SOCKET_ERROR)
  {
    // print("Create stream socket failed\n");
    return(-1);
  }


  // max. send and receive buffer for fast ethernet transfer
  #if 1
  {
    #define SEND_BUFFER_SIZE    (0xFFFF) // max window size
    #define RECEIVE_BUFFER_SIZE (0xFFFF) // max window size

    I32 opt_value;

    // send buffer size
    opt_value = SEND_BUFFER_SIZE;
    Error = setsockopt(sock, SOL_TCP, OPT_TBSIZE, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
     print("Could not set socket option: OPT_TBSIZE!\n");
    }

    // receive buffer size
    opt_value = RECEIVE_BUFFER_SIZE;
    Error = setsockopt(sock, SOL_TCP, OPT_RBSIZE, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
     print("Could not set socket option: OPT_RBSIZE!\n");
    }
  }
  #endif

  #if 0
  {
    /* send data immediately */
    I32 opt_value = 1;

    Error = setsockopt(sock, SOL_TCP, OPT_SEND_PUSH, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
      print("ERROR: OPT_SEND_PUSH!\n");
    }
  }
  #endif

  #if 1
  {
    /* send data immediately (VCRT_MSG_NONBLOCK) */
    I32 opt_value = 1;

    Error = setsockopt(sock, SOL_TCP, OPT_SEND_NOWAIT, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
      print("ERROR: OPT_SEND_NOWAIT!\n");
    }
  }
  #endif




  /* receive immediatly the data */
  /*
  error = setsockopt(sock, 0, OPT_RECEIVE_NOWAIT, &opt_value, opt_length);
  if (error != VCRT_OK)
  {
    print("Could not set socket option: OPT_RECEIVE_NOWAIT!\n");
  }
  */

  /* connection between free port and selected SERVER_PORT. Now SERVER_PORT is free again for new connections. */
  error = bind(sock, &laddr, sizeof(laddr));

  if (error != VCRT_OK)
  {
    // print("Stream bind failed - 0x%x\n", error);
    return(-2);
  }

  /* activate sock for TCP/IP connections */
  #ifdef TCP_IP_SOCKET
  error = listen(sock, 0);
  #endif

  if (error != VCRT_OK)
  {
    // print("Listen failed - 0x%x\n", error);
    return(-3);
  }

  // print("\nFast Ethernet port %d\n\n", laddr.sin_port);

  *NewSock = sock;
  return(0);
}

/************************************************************************************************************************/

I32 TCP_Connection (U32 timeout, U32 *NewSock)
{
   /*
   timeout:
   -  0 means indefinitely
   - -1 means do not block
   -  else waiting time in ms

   return
   -  socket
   - -1 if no connection
   - -2 time out
   */

   *NewSock = VCRT_selectall(timeout);
   // *NewSock = VCRT_selectset(*NewSock, 1, timeout);

   if (*NewSock == VCRT_SOCKET_ERROR) return(-1);
   if (*NewSock <=                 0) return(-2); // timeout

   return(0);
}

/************************************************************************************************************************/

I32 TCP_Accept (U32 Sock, U32 *FromSock)
{
   uint_16  rlen;
   sockaddr_in  raddr;

   /* Connection requested; accept it */
   rlen = sizeof(raddr);

   *FromSock = accept(Sock, &raddr, &rlen);

   if (*FromSock == VCRT_SOCKET_ERROR)
   {
     return(-1);
   }

   return(0);
}

/************************************************************************************************************************/

I32 TCP_Send (U32 Sock, char *SendData, I32 SendBytes)
{
  I32 result;

#if 0
  {
   /* sends immediatly the data */
   I32 opt_value  = 1;
   I32 opt_length = sizeof(unsigned int);
   I32 error;
   error = setsockopt(Sock, SOL_TCP, OPT_SEND_PUSH, &opt_value, opt_length);
   if (error != VCRT_OK)
   {
   print("ERROR: OPT_SEND_PUSH!\n");
   }
 }
#endif

  if(getvar(ETHLINK))
  {
    result = send(Sock, (char *)SendData, SendBytes, 0);
  }
  else
  {
    result = VCRT_ERROR;
  }

  if(result==VCRT_ERROR)
  {
    return(-1);
  }

  return(result);
}

/************************************************************************************************************************/

I32 TCP_Receive (U32 Sock, char *ReceiveData, I32 ReceiveBytes)
{
   return( recv(Sock, (char *)ReceiveData, ReceiveBytes, 0) );
}

/************************************************************************************************************************/

void TCP_Close (U32 Sock)
{
   shutdown(Sock, FLAG_CLOSE_TX);
}

/************************************************************************************************************************/


/************************************************************************************************************************/
/************************************************************************************************************************/
/************************************************************************************************************************/

